from typing import TYPE_CHECKING
import asyncio
from copy import deepcopy
from shimoku.api.utils import change_data_set_name_with_report, revert_uuids_from_dict
from shimoku.code_gen.business_code_gen.apps_code_gen. \
    data_sets_code_gen.code_gen_from_data_sets import code_gen_read_csv_from_data_set, get_linked_data_set_info
from shimoku.code_gen.utils_code_gen import code_gen_from_list

if TYPE_CHECKING:
    from shimoku.api.resources.report import Report
    from shimoku.code_gen.business_code_gen.apps_code_gen.code_gen_from_apps import AppCodeGen


# Todo needs to have the correct names for the columns
async def code_gen_from_annotated_echart(
        self: 'AppCodeGen', report: 'Report', properties: dict
) -> list[str]:
    """ Generate code for an echarts report.
    :param report: report to generate code from
    :param report_params: parameters of the report
    :param properties: properties of the report
    :return: list of code lines """
    echart_options = deepcopy(properties['option'])
    rds_ids_in_order = revert_uuids_from_dict(echart_options)
    data_set_ids, _ = await get_linked_data_set_info(self, report, rds_ids_in_order)
    data_sets = await asyncio.gather(*[self._app.get_data_set(ds_id) for ds_id in data_set_ids])
    if data_sets[0]['columns']:
        rev_data_set_columns = [{v: k for k, v in data_set['columns'].items()} for data_set in data_sets]
    else:
        rev_data_set_columns = [{"dateField1": "dateField1", "intField1": "intField1"} for _ in data_sets]
    data_set_names = [change_data_set_name_with_report(data_set, report) for data_set in data_sets]
    data_args = await asyncio.gather(*[code_gen_read_csv_from_data_set(data_set, name)
                                       for data_set, name in zip(data_sets, data_set_names)])
    data_args = [data_arg for data_arg in data_args if data_arg is not None]
    slider_properties = properties.get('slider') or {}
    marks = slider_properties.pop('marks') if 'marks' in slider_properties else None

    slider_params = []
    if slider_properties:
        slider_params.append(f'    slider_config={slider_properties},')
    if marks:
        slider_params.append(f'    slider_marks={[(mark["label"], mark["value"]) for mark in marks]},')
    y_code_lines = code_gen_from_list([rev_data_set_columns[i]["intField1"] for i in range(len(data_args))], 4)
    return [
        'shimoku_client.plt.annotated_chart(',
        f'    data=[{", ".join(data_args)}],',
        f'    x="{rev_data_set_columns[0]["dateField1"]}",',
        f'    y={y_code_lines[0][4:]}',
        *y_code_lines[1:],
        *slider_params,
        *self.code_gen_report_params(report.cascade_to_dict()),
        ')'
    ]



import unittest
from test_plot_api import main, shimoku_client as tpa_shimoku_client
import test_plot_api
from utils import initiate_shimoku
from shimoku import Client
import tempfile
import subprocess
import shutil
import json
import datetime as dt
from deepdiff import DeepDiff
from os import getenv

mock = getenv('MOCK') == 'TRUE'


def get_report_data_set_info(
    r_hash: str, report_data_set: dict, data_sets_mapping: dict[str, str]
) -> dict:
    report_data_set.pop('id')
    report_id = report_data_set.pop('reportId')
    report_data_set['dataSetId'] = data_sets_mapping[report_data_set['dataSetId']].replace(report_id, r_hash)
    return report_data_set


def get_report_data_set_info_dict(
    r_hash: str, _dict: dict, report_data_sets: dict[str, dict], data_sets_mapping: dict[str, str]
):
    for k, v in _dict.items():
        if isinstance(v, dict):
            get_report_data_set_info_dict(r_hash, v, report_data_sets, data_sets_mapping)
        elif isinstance(v, list):
            get_report_data_set_info_list(r_hash, v, report_data_sets, data_sets_mapping)
        elif isinstance(v, str) and v.startswith('#{') and v.endswith('}'):
            uuid = v[2:-1]
            _dict[k] = get_report_data_set_info(r_hash, report_data_sets[uuid], data_sets_mapping)


def get_report_data_set_info_list(
    r_hash: str, _list: list, report_data_sets: dict[str, dict], data_sets_mapping: dict[str, str]
):
    for i, v in enumerate(_list):
        if isinstance(v, dict):
            get_report_data_set_info_dict(r_hash, v, report_data_sets, data_sets_mapping)
        elif isinstance(v, list):
            get_report_data_set_info_list(r_hash, v, report_data_sets, data_sets_mapping)
        elif isinstance(v, str) and v.startswith('#{') and v.endswith('}'):
            uuid = v[2:-1]
            _list[i] = get_report_data_set_info(r_hash, report_data_sets[uuid], data_sets_mapping)


def clear_workspace(shimoku_client):
    for menu_path in shimoku_client.workspaces.get_workspace_menu_paths(shimoku_client.workspace_id):
        shimoku_client.menu_paths.delete_all_menu_path_activities(uuid=menu_path['id'])
    shimoku_client.workspaces.delete_all_workspace_menu_paths(shimoku_client.workspace_id)
    shimoku_client.workspaces.delete_all_workspace_boards(shimoku_client.workspace_id)


def change_ids_in_tabs_or_modal(component, components_by_id):
    if 'tabs' in component['properties']:
        for tab_name, tab_dict in component['properties']['tabs'].items():
            component['properties']['tabs'][tab_name]['reportIds'] = sorted([
                components_by_id[r_id]['properties']['hash'] for r_id in tab_dict['reportIds']
            ])
    if 'reportIds' in component['properties']:
        component['properties']['reportIds'] = sorted([
            components_by_id[r_id]['properties']['hash'] for r_id in component['properties']['reportIds']
        ])


def handle_events(events, components_by_id):
    on_click_events = events.get('onClick', [])
    for event in on_click_events:
        if 'modalId' in event['params']:
            event['params']['modalId'] = components_by_id[event['params']['modalId']]['properties']['hash']


def handle_bentobox(component, current_bentobox):
    if not component.get('bentobox'):
        current_bentobox['id_to_replace'] = ''
        return
    bentobox_id = component['bentobox']['bentoboxId']

    if bentobox_id != current_bentobox['id_to_replace']:
        current_bentobox['id_to_replace'] = bentobox_id
        current_bentobox['bentoboxId'] = f'_{component["order"]}'
        current_bentobox['bentoboxOrder'] = component['order']

    component['bentobox']['bentoboxId'] = current_bentobox['bentoboxId']
    component['bentobox']['bentoboxOrder'] = current_bentobox['bentoboxOrder']


def handle_component(shimoku_client, component, mpath_components_by_id, data_sets_mapping, current_bentobox):
    data_set_links = shimoku_client.components.get_component_data_set_links(component['id'])
    get_report_data_set_info_dict(
        component['properties']['hash'], component['properties'],
        {rds['id']: rds for rds in data_set_links}, data_sets_mapping
    )
    handle_bentobox(component, current_bentobox)
    if 'events' in component['properties']:
        handle_events(component['properties']['events'], mpath_components_by_id)
    if component['reportType'] in ['MODAL', 'TABS']:
        change_ids_in_tabs_or_modal(component, mpath_components_by_id)
    elif component['reportType'] == 'FILTERDATASET':
        fds_mapping = component['properties']['mapping']
        fds_mapping[0]['id'] = data_sets_mapping[fds_mapping[0]['id']]


def handle_menu_path(shimoku_client, menu_path, components, data_sets_mapping, data_by_data_set):
    shimoku_client.set_menu_path(menu_path['name'], dont_add_to_dashboard=True)
    mpath_components = sorted(shimoku_client.menu_paths.get_menu_path_components(menu_path['id']),
                              key=lambda x: x['properties']['hash'])
    components.extend(mpath_components)
    components_hash_by_id = {c['id']: c['properties']['hash'] for c in mpath_components}
    data_sets = shimoku_client.menu_paths.get_menu_path_data_sets(menu_path['id'])
    for data_set in data_sets:
        for comp_id in components_hash_by_id.keys():
            if comp_id in data_set['name']:
                data_set['name'] = data_set['name'].replace(comp_id, components_hash_by_id[comp_id])
        data_sets_mapping[data_set['id']] = data_set['name']

    for data_set in sorted(data_sets, key=lambda x: x['name']):
        data_by_data_set[data_set['name']] = shimoku_client.data.get_data_from_data_set(data_set['id'])

    mpath_components_by_id = {c['id']: c for c in mpath_components}
    current_bentobox = {'id_to_replace': '', 'bentoboxId': '', 'bentoboxOrder': ''}
    for component in mpath_components:
        handle_component(shimoku_client, component, mpath_components_by_id, data_sets_mapping, current_bentobox)


def get_diff_percentage(items1, items2):
    return sum(DeepDiff(c1, c2) != {} for c1, c2 in zip(items1, items2))/len(items1)


def get_workspace_contents(shimoku_client):
    boards = sorted(
        shimoku_client.workspaces.get_workspace_boards(shimoku_client.workspace_id),
        key=lambda x: x['order']
    )
    menu_paths = sorted(
        shimoku_client.workspaces.get_workspace_menu_paths(
            shimoku_client.workspace_id
        ),
        key=lambda x: x['order']
    )
    components = []
    data_sets_mapping = {}
    data_by_data_set = {}
    for menu_path in menu_paths:
        handle_menu_path(shimoku_client, menu_path, components, data_sets_mapping, data_by_data_set)
        shimoku_client.pop_out_of_menu_path()
    _all = [*boards, *menu_paths, *components, *[datum for data in data_by_data_set.values() for datum in data]]
    for element in _all:
        element.pop('id')
    return boards, menu_paths, components, data_by_data_set


def create_bar_chart(shimoku_client):
    data = [{'date': dt.date(2021, 1, 1), 'x': 50000000, 'y': 5},
            {'date': dt.date(2021, 1, 2), 'x': 60000000, 'y': 5},
            {'date': dt.date(2021, 1, 3), 'x': 40000000, 'y': 5},
            {'date': dt.date(2021, 1, 4), 'x': 70000000, 'y': 5},
            {'date': dt.date(2021, 1, 5), 'x': 30000000, 'y': 5}]
    shimoku_client.plt.bar(
        data=data, x='date', y=['x', 'y'],
        x_axis_name='Date', y_axis_name='Revenue',
        order=0, rows_size=2,
        cols_size=12,
    )


class TestCodeGen(unittest.TestCase):

    def test_code_gen(self):
        if not tpa_shimoku_client.playground or mock:
            # It would take too long to complete the test
            return
        tpa_shimoku_client.set_workspace()
        clear_workspace(tpa_shimoku_client)
        main()
        tpa_shimoku_client.pop_out_of_board()
        tpa_shimoku_client.disable_caching()
        boards, menu_paths, components, data = get_workspace_contents(tpa_shimoku_client)
        temp_dir = tempfile.mkdtemp()
        subprocess.run([
            'shimoku', 'persist', 'generate-code',
            '--output-path', temp_dir,
            '--universe-id', 'local',
            '--workspace', 'local',
        ])
        clear_workspace(tpa_shimoku_client)
        subprocess.run(['python', f'{temp_dir}/execute_workspace_local.py'], check=True)
        (first_generation_boards,
         first_generation_menu_paths,
         first_generation_components,
         first_generation_data) = get_workspace_contents(tpa_shimoku_client)
        shutil.rmtree(temp_dir)
        temp_dir = tempfile.mkdtemp()
        subprocess.run([
            'shimoku', 'persist', 'generate-code',
            '--output-path', temp_dir,
            '--universe-id', 'local',
            '--workspace', 'local',
        ])
        subprocess.run(['python', f'{temp_dir}/execute_workspace_local.py'], check=True)
        (second_generation_boards,
         second_generation_menu_paths,
         second_generation_components,
         second_generation_data) = get_workspace_contents(tpa_shimoku_client)
        clear_workspace(tpa_shimoku_client)
        tpa_shimoku_client.run()
        shutil.rmtree(temp_dir)
        results = {
            'orig-first': (get_diff_percentage(boards, first_generation_boards),
                           get_diff_percentage(menu_paths, first_generation_menu_paths),
                           get_diff_percentage(components, first_generation_components),
                           get_diff_percentage(data, first_generation_data)),
            'orig-second': (get_diff_percentage(boards, second_generation_boards),
                            get_diff_percentage(menu_paths, second_generation_menu_paths),
                            get_diff_percentage(components, second_generation_components),
                            get_diff_percentage(data, second_generation_data)),
            'first-second': (get_diff_percentage(first_generation_boards, second_generation_boards),
                             get_diff_percentage(first_generation_menu_paths, second_generation_menu_paths),
                             get_diff_percentage(first_generation_components, second_generation_components),
                             get_diff_percentage(first_generation_data, second_generation_data)),
        }
        print(json.dumps(results, indent=4))
        test_plot_api.set_shimoku_client()
        assert all(all([value == 0 for value in results_list]) for results_list in results.values())

    def test_commit(self):
        if tpa_shimoku_client.playground or mock:
            # Test commit only in cloud
            return
        local_shimoku_client = Client(verbosity='INFO')
        local_shimoku_client.disable_caching()
        local_shimoku_client.set_workspace()
        clear_workspace(local_shimoku_client)
        local_shimoku_client.set_board('Commit-board')
        local_shimoku_client.set_menu_path('Commit-menu-path')
        create_bar_chart(local_shimoku_client)
        local_boards, local_menu_paths, local_components, local_data = get_workspace_contents(local_shimoku_client)
        clear_workspace(tpa_shimoku_client)
        subprocess.run([
            'shimoku', 'persist', 'commit',
            '--universe-id', 'local',
            '--workspace', 'local',
            '--target-access-token', tpa_shimoku_client.access_token,
            '--target-universe-id', tpa_shimoku_client.universe_id,
            '--target-workspace', tpa_shimoku_client.workspace_id,
            '--target-environment', tpa_shimoku_client.environment,
        ])
        tpa_shimoku_client.disable_caching()
        tpa_shimoku_client.set_workspace(tpa_shimoku_client.workspace_id)
        boards, menu_paths, components, data = get_workspace_contents(tpa_shimoku_client)
        results = {
            'boards': (get_diff_percentage(local_boards, boards)),
            'menu_paths': (get_diff_percentage(local_menu_paths, menu_paths)),
            'components': (get_diff_percentage(local_components, components)),
            'data': (get_diff_percentage(local_data, data)),
        }
        print(json.dumps(results, indent=4))
        print(local_menu_paths)
        print(menu_paths)
        assert all([value == 0 for value in results.values()])

        test_plot_api.set_shimoku_client()

    def test_pull(self):
        if tpa_shimoku_client.playground or mock:
            # Test pull only in cloud
            return
        tpa_shimoku_client.set_workspace(tpa_shimoku_client.workspace_id)
        clear_workspace(tpa_shimoku_client)
        tpa_shimoku_client.set_board('Pull-board')
        tpa_shimoku_client.set_menu_path('Pull-menu-path')
        create_bar_chart(tpa_shimoku_client)
        tpa_boards, tpa_menu_paths, tpa_components, tpa_data = get_workspace_contents(tpa_shimoku_client)

        local_shimoku_client = Client(verbosity='INFO')
        local_shimoku_client.disable_caching()
        local_shimoku_client.set_workspace()
        clear_workspace(local_shimoku_client)
        subprocess.run([
            'shimoku', 'persist', 'pull',
            '--universe-id', 'local',
            '--workspace', 'local',
            '--origin-access-token', tpa_shimoku_client.access_token,
            '--origin-universe-id', tpa_shimoku_client.universe_id,
            '--origin-workspace', tpa_shimoku_client.workspace_id,
            '--origin-environment', tpa_shimoku_client.environment,
        ])
        local_boards, local_menu_paths, local_components, local_data = get_workspace_contents(local_shimoku_client)

        clear_workspace(tpa_shimoku_client)
        clear_workspace(local_shimoku_client)

        results = {
            'boards': (get_diff_percentage(tpa_boards, local_boards)),
            'menu_paths': (get_diff_percentage(tpa_menu_paths, local_menu_paths)),
            'components': (get_diff_percentage(tpa_components, local_components)),
            'data': (get_diff_percentage(tpa_data, local_data)),
        }
        print(json.dumps(results, indent=4))
        print(local_menu_paths)
        print(tpa_menu_paths)
        assert all([value == 0 for value in results.values()])

        test_plot_api.set_shimoku_client()



import pandas as pd
import asyncio
import numpy as np

from copy import deepcopy

from pandas import DataFrame
from math import ceil
from shimoku.async_execution_pool import AutoAsyncExecutionPool, add_general_async_group

from typing import Optional, Union

from shimoku.utils import EventType

from shimoku.api.resources.app import App
from shimoku.api.resources.business import Business
from shimoku.api.resources.report import Report
from shimoku.api.resources.data_set import DataSet, Mapping, convert_input_data_to_db_items
from shimoku.api.resources.reports.filter_data_set import FilterDataSet
from shimoku.api.resources.reports.tabs_group import TabsGroup
from shimoku.api.resources.reports.modal import Modal
from shimoku.api.resources.reports.charts.indicator import Indicator
from shimoku.api.resources.reports.charts.echart import EChart
from shimoku.api.resources.reports.charts.table import Table, interpret_label_info
from shimoku.api.resources.reports.charts.annotated_chart import AnnotatedEChart
from shimoku.api.resources.reports.charts.html import HTML
from shimoku.api.resources.reports.charts.button import Button
from shimoku.api.resources.reports.charts.input_form import InputForm
from shimoku.api.resources.reports.charts.iframe import IFrame
from shimoku.plt.EChart_definitions.line import (line_chart, area_chart, stacked_area_chart,
                                                 predictive_line_chart, marked_line_chart,
                                                 segmented_area_chart,
                                                 line_with_confidence_area_chart, segmented_line_chart)
from shimoku.plt.EChart_definitions.bar import (bar_chart, stacked_bar_chart, horizontal_bar_chart,
                                                stacked_horizontal_bar_chart, zero_centered_bar_chart)
from shimoku.plt.EChart_definitions.scatter import scatter_chart, scatter_with_effect_chart
from shimoku.plt.EChart_definitions.funnel import funnel_chart
from shimoku.plt.EChart_definitions.tree import tree_chart
from shimoku.plt.EChart_definitions.radar import radar_chart
from shimoku.plt.EChart_definitions.pie import pie_chart, doughnut_chart, rose_chart
from shimoku.plt.EChart_definitions.gauge import speed_gauge_chart
from shimoku.plt.EChart_definitions.shimoku_gauge import shimoku_gauge_chart, shimoku_gauges_group
from shimoku.plt.EChart_definitions.sunburst import sunburst_chart
from shimoku.plt.EChart_definitions.treemap import treemap_chart
from shimoku.plt.EChart_definitions.sankey import sankey_chart
from shimoku.plt.EChart_definitions.heatmap import heatmap_chart
from shimoku.plt.EChart_definitions.gauge_indicator import gauge_indicator
from shimoku.plt.EChart_definitions.top_bottom import top_bottom_area_charts, top_bottom_line_charts
from shimoku.plt.EChart_definitions.waterfall import waterfall_chart
from shimoku.plt.EChart_definitions.line_and_bar import line_and_bar_charts
from shimoku.plt.bentobox_charts import chart_and_modal_button, infographics_text_bubble, \
    indicators_with_header, chart_and_indicators, line_with_summary

from shimoku.exceptions import TabsError, ModalError, DataError, BentoboxError
from shimoku.plt.utils import (deep_update, get_uuids_from_dict,
                               get_data_references_from_dict, validate_data_is_pandarable, \
                               add_sorting_to_df, transform_dict_js_to_py, retrieve_data_from_options,
                               validate_input_form_data, create_normalized_name, convert_data_and_get_series_name)

import logging
from shimoku.execution_logger import log_error, ClassWithLogging

logger = logging.getLogger(__name__)


class PlotLayer(ClassWithLogging):
    """ 
    This class is a high level abstraction of the API, it is used to create components and data sets easily.
    """
    
    _module_logger = logger
    _use_info_logging = True

    def __init__(self, async_pool: AutoAsyncExecutionPool, app: Optional[App], reuse_data_sets: bool = False):
        self._app = app
        self._business: Business = app.parent if app else None
        self._current_path: Optional[str] = None
        self._current_tabs_group: Optional[TabsGroup] = None
        self._current_tab: Optional[str] = None
        self._current_modal: Optional[Modal] = None
        self._bentobox_data: dict = {}
        self.reuse_data_sets: bool = reuse_data_sets
        self._delete_data_set_lock = None
        self._shared_data_map: dict[str, dict[str, tuple[Mapping, DataSet, dict]]] = {}
        self._shared_data: dict[str, any] = {}
        self._execution_path_orders: list[str] = []
        self._async_pool = async_pool

    async def _create_event(self, event_type: EventType, content: dict, resource_id: Optional[str] = None):
        """ Create an event.
        :param event_type: the type of the event
        :param content: the content of the event
        """
        if self._app.api_client.playground:
            await self._business.create_event(event_type, content, resource_id)

    def clear_context(self):
        if self._bentobox_data:
            self._bentobox_data = {}
            logger.info('Popped out of bentobox')
        if self._current_tabs_group:
            self._current_tabs_group = None
            self._current_tab = None
            logger.info('Popped out of tabs group')
        if self._current_modal:
            self._current_modal = None
            logger.info('Popped out of modal')

    def _check_for_conflicts(self, async_pool: AutoAsyncExecutionPool, order: int):
        """
        Check if there are charts with the same order.
        :param order: the order of the chart
        """
        r_hash = self._get_chart_hash(order)
        free_context: dict = async_pool.free_context
        if 'list_for_conflicts' not in free_context:
            free_context['list_for_conflicts'] = []

        if r_hash in free_context['list_for_conflicts']:
            async_pool.clear()
            self.clear_context()
            log_error(logger, 'Chart order collision, two charts with the same order can not be executed '
                              'at the same time', RuntimeError)

        free_context['list_for_conflicts'].append(r_hash)

    def _check_before_async_execution(self, async_pool: AutoAsyncExecutionPool, func: callable, *args, **kwargs):
        """ Check everything is correct before executing the task pool """
        if 'order' in kwargs:
            self._check_for_conflicts(async_pool, kwargs['order'])

    def raise_if_cant_change_path(self):
        """Raise an error if a tabs group or a modal is already open. """
        if self._current_tabs_group:
            log_error(logger, 'Cannot change path while a tabs group is open', TabsError)
        if self._current_modal:
            log_error(logger, 'Cannot change path while a modal is open', ModalError)
        if self._bentobox_data:
            log_error(logger, 'Cannot change path while a bentobox is open', BentoboxError)

    def get_shared_data_names(self) -> list[str]:
        """ Get the names of the shared data """
        return list(self._shared_data_map.keys())

    def change_path(self, path: str):
        """ Change the current path """
        self.raise_if_cant_change_path()
        self._current_path = path
        if path and path not in self._execution_path_orders:
            self._execution_path_orders.append(path)

    def _get_hash_for_container(self, name: str):
        """ Get the hash for a container """
        return (f'{self._current_path}_' if self._current_path else '') + name

    async def clear_menu_path(self):
        """
        Clear the current path or a subpath
        """
        reports = await self._app.get_reports()
        touched_data_set_ids = []
        if self._current_path is not None:
            path = create_normalized_name(self._current_path)
            reports = [report for report in reports
                       if report['path'] and create_normalized_name(report['path']) == path]
        else:
            touched_data_set_ids = [ds['id'] for ds in await self._app.get_data_sets()]

        containers: list = [report for report in reports if report.report_type in ['TABS', 'MODAL']]

        for container in containers:
            container.clear_content()

        rds = await asyncio.gather(*[report.get_report_data_sets() for report in reports])
        touched_data_set_ids.extend([rd['dataSetId'] for rds in rds for rd in rds])

        await asyncio.gather(*[self._app.delete_report(report['id']) for report in reports])
        logger.info(f'Deleted {len(reports)} components')
        await self._app.delete_unused_data_sets(log=True, data_set_ids=touched_data_set_ids)

    async def delete_chart_by_order(self, order: int):
        """
        Delete a report by order and context
        """
        r_hash, report = await self._get_chart_report(order, Table, create_if_not_exists=False)
        if not report:
            log_error(logger, f'No chart found with order {order}', RuntimeError)
        await self._app.delete_report(r_hash=r_hash)

    async def delete_tabs_group(self, name: str):
        """ Delete a tabs group """
        r_hash = self._get_hash_for_container(name)
        tabs_group = await self._app.get_report(r_hash=r_hash)
        if not tabs_group:
            log_error(logger, f'No tabs group found with name {name}', TabsError)
        await self._app.delete_report(r_hash=r_hash)

    async def delete_modal(self, name: str):
        """ Delete a modal """
        r_hash = self._get_hash_for_container(name)
        modal = await self._app.get_report(r_hash=r_hash)
        if not modal:
            log_error(logger, f'No modal found with name {name}', ModalError)
        await self._app.delete_report(r_hash=r_hash)

    def set_bentobox(self, cols_size: int, rows_size: int):
        """ Start using a bentobox, the id and the order will be set when the bentobox is used for the first time
        :param cols_size: the number of columns in the bentobox
        :param rows_size: the number of rows in the bentobox"""
        self._bentobox_data: dict = {
            'bentoboxId': None,
            'bentoboxOrder': None,
            'bentoboxSizeColumns': cols_size,
            'bentoboxSizeRows': rows_size,
        }

    def pop_out_of_bentobox(self):
        """ Stop using a bentobox """
        self._bentobox_data = {}

    def _get_bentobox_data(self, order: int) -> dict:
        """ Get the bentobox data """
        if not self._bentobox_data:
            return {}
        if not self._bentobox_data['bentoboxId']:
            if not self._bentobox_data['bentoboxId']:
                self._bentobox_data.update({
                    'bentoboxId': '_' + str(order),
                    'bentoboxOrder': order,
                })
        return self._bentobox_data

    async def _update_containers(self):
        """ Update the reports that act as containers"""
        reports = await self._app.get_reports()
        containers: list[Report] = [report for report in reports if report.report_type in ['TABS', 'MODAL']]
        await asyncio.gather(*[container.update() for container in containers])
        logger.info('Updated tab groups and modals')

    async def set_tabs_index(
            self, tabs_index: tuple[str, str], order: Optional[int] = None,
            parent_tabs_index: Optional[tuple[str, str]] = None, padding: Optional[str] = None,
            cols_size: Optional[int] = None, rows_size: Optional[int] = None,
            just_labels: Optional[bool] = None, sticky: Optional[bool] = None
    ):
        """ Set the current tabs index.
        :param tabs_index: the index of the tabs group
        :param order: the order of the tabs group in the dashboard
        :param parent_tabs_index: the index of the parent tabs group
        :param cols_size: the size of the columns in the tabs group
        :param padding: the padding of the tabs group
        :param rows_size: the size of the rows in the tabs group
        :param just_labels: whether to show just the labels in the tabs group
        :param sticky: whether to make the tabs group sticky
        """
        r_hash = self._get_hash_for_container(tabs_index[0])
        tabs_group: Optional[TabsGroup] = await self._app.get_report(r_hash=r_hash)
        params = {'properties': {}}
        if cols_size:
            params['sizeColumns'] = cols_size
        if padding:
            params['sizePadding'] = padding
        if rows_size:
            params['sizeRows'] = rows_size
        if isinstance(just_labels, bool):
            params['properties']['variant'] = 'solidRounded' if just_labels else 'enclosedSolidRounded'
        if isinstance(sticky, bool):
            params['properties']['sticky'] = sticky
        bentobox_data = self._get_bentobox_data(order)
        if bentobox_data:
            params['bentobox'] = bentobox_data
        if self._current_path:
            params['path'] = self._current_path

        if not tabs_group:
            if not isinstance(order, int):
                log_error(logger, f'Cannot create tabs group {tabs_index[0]} without order', TabsError)
            tabs_group: TabsGroup = await self._app.create_report(TabsGroup, r_hash=r_hash, order=order, **params)
            logger.info(f'Created tabs group {tabs_index[0]} with id {tabs_group["id"]}')
        elif order:
            await self._app.update_report(r_hash=r_hash, order=order, **params)

        if parent_tabs_index:
            p_hash = self._get_hash_for_container(parent_tabs_index[0])
            parent_tabs_group: Optional[TabsGroup] = await self._app.get_report(r_hash=p_hash)
            if not parent_tabs_group:
                log_error(logger, f'No tabs group found with name {parent_tabs_index[0]}', TabsError)
            if parent_tabs_group['id'] == tabs_group['id']:
                log_error(logger, f'Cannot include tabs group in itself', TabsError)
            if self._current_modal:
                log_error(logger, f'Cannot include a tabs group in a modal and in another tabs group', TabsError)
            if not parent_tabs_group.has_report(tabs_group):
                parent_tabs_group.add_report(tab=parent_tabs_index[1], report=tabs_group)
                logger.info(f'Included tabs group {tabs_index[0]} in tabs group {parent_tabs_index[0]}')

        elif self._current_modal and not self._current_modal.has_report(tabs_group):
            self._current_modal.add_report(tabs_group)
            logger.info(f'Included tabs group {tabs_index[0]} in modal {str(self._current_modal)}')

        self._current_tabs_group = tabs_group
        tabs_group.add_tab(tabs_index[1])
        self._current_tab = tabs_index[1]

        if 'update_containers' not in self._async_pool.ending_tasks:
            self._async_pool.ending_tasks['update_containers'] = self._update_containers()

    def pop_out_of_tabs_group(self):
        """
        Pop the current tabs index.
        """
        if not self._current_tabs_group:
            log_error(logger, f'No tabs group to pop out of', TabsError)

        self._current_tabs_group = None
        self._current_tab = None

    def change_current_tab(self, tab: str):
        """
        Change the current tab.
        :param tab: the name of the tab
        """
        if not self._current_tabs_group:
            log_error(logger, f'No tabs group to change tab', TabsError)

        self._current_tabs_group.add_tab(tab)
        self._current_tab = tab

        if 'update_containers' not in self._async_pool.ending_tasks:
            self._async_pool.ending_tasks['update_containers'] = self._update_containers()

    async def _get_modal(
        self, modal_name: str,
        width: Optional[int] = None,
        height: Optional[int] = None,
        open_by_default: Optional[bool] = None
    ) -> Modal:
        r_hash = self._get_hash_for_container(modal_name)
        modal: Optional[Modal] = await self._app.get_report(r_hash=r_hash)
        if not modal:
            modal: Report = await self._app.create_report(
                Modal, r_hash=r_hash, path=self._current_path,
                properties={'width': width, 'height': height, 'open': open_by_default})
            logger.info(f'Created modal {modal_name} with id {modal["id"]}')
            return modal
        properties = {}
        if width:
            properties['width'] = width
        if height:
            properties['height'] = height
        if open_by_default:
            properties['open'] = open_by_default
        if properties:
            await self._app.update_report(r_hash=r_hash, properties=properties, path=self._current_path)
            logger.info(f'Updated modal {modal_name} with id {modal["id"]}')
        return modal

    async def set_modal(
        self, modal_name: str,
        width: Optional[int] = None,
        height: Optional[int] = None,
        open_by_default: Optional[bool] = None
    ) -> None:
        """ Set the current modal.
        :param modal_name: the name of the modal
        :param width: the width of the modal
        :param height: the height of the modal
        :param open_by_default: whether the modal is open by default
        """
        if self._current_tabs_group:
            log_error(logger, f'Cannot set a modal in a tabs group, pop out of the tabs group first', ModalError)

        modal = await self._get_modal(modal_name, width, height, open_by_default)
        self._current_modal = modal

        if 'update_containers' not in self._async_pool.ending_tasks:
            self._async_pool.ending_tasks['update_containers'] = self._update_containers()

    def pop_out_of_modal(self):
        """ Pop the current modal. """
        if not self._current_modal:
            log_error(logger, f'No modal to pop out of', ModalError)

        if self._current_tabs_group:
            log_error(logger,
                      f'Cannot pop out of a modal when in a tabs group, pop out of the tabs group first', ModalError)
        if self._bentobox_data:
            log_error(logger, f'Cannot pop out of a modal when in a bentobox, close the bentobox first', ModalError)

        self._current_modal = None

    def _get_chart_hash(self, order: int) -> str:
        r_hash = f'{order}'
        if self._current_tabs_group:
            r_hash = f'{self._current_tabs_group["properties"]["hash"]}_{self._current_tab}_{order}'
        elif self._current_modal:
            r_hash = f'{self._current_modal["properties"]["hash"]}_{order}'
        elif self._current_path:
            r_hash = f'{self._current_path}_{order}'
        return r_hash

    async def _get_chart_report(
            self, order: int, chart_class: type[Report],
            create_if_not_exists: bool = True
    ) -> tuple[str, Optional[Report]]:
        """ Get the chart report.
        :param order: the order of the chart
        :param chart_class: the chart class
        :param create_if_not_exists: whether to create the chart if it doesn't exist
        """
        r_hash = self._get_chart_hash(order)
        report = await self._app.get_report(r_hash=r_hash)

        if not report:
            if create_if_not_exists:
                report = await self._app.create_report(chart_class, r_hash=r_hash, order=order)
                logger.info(f'Created {chart_class.__name__} with id {report["id"]}')
        elif report.report_type != chart_class.report_type:
            await report.change_report_type(chart_class)

        return r_hash, report

    async def _try_to_reuse_data(
            self, data_set: DataSet, data: Union[list, pd.DataFrame]
    ) -> dict[str, tuple[Mapping, DataSet, dict]]:
        """ Try to reuse the data set.
        :param data_set: the data set
        :param data: the data
        """
        df = validate_data_is_pandarable(data)
        aux_data_point = await data_set.get_one_data_point()
        if not aux_data_point:
            log_error(logger, f'Cannot reuse data set {str(data_set)} because the data set is empty', DataError)

        mappings = [col for col in aux_data_point if aux_data_point[col] is not None]
        df, sort = add_sorting_to_df(df) if 'orderField1' in mappings else (df, None)

        converted_data_points = convert_input_data_to_db_items(df, sort=sort)
        for mapping, v in converted_data_points[0].items():
            if mapping not in mappings:
                log_error(logger, f'Cannot reuse data set {str(data_set)} because the data provided '
                                  f'is not consistent with the data set', DataError)

        first_df_item = df.iloc[0].to_dict()
        return {col: (mapping, data_set, sort) for col, mapping
                in zip(first_df_item.keys(), converted_data_points[0].keys())
                if col != 'sort_values'}

    async def _delete_data_set_if_exists(self, data_set_name: str) -> None:
        """ Delete the data set if it exists.
        :param data_set_name: the name of the data set
        """
        if await self._app.force_delete_data_set(name=data_set_name):
            logger.info(f'Deleted data set with name {data_set_name}')

    async def _create_data_set_from_df(
            self, data_set_name: str, df: pd.DataFrame, sort: Optional[dict] = None
    ) -> dict[str, tuple[Mapping, DataSet, dict]]:
        """ Get the data mapping to tuple.
        :param df: the data frame
        :return: the data mapping to tuple
        """
        df, sort = add_sorting_to_df(df, sort)
        columns = df.columns.tolist()
        data_mapping_to_tuple = {}
        (mapping, data_set, res_sort) = await self._app.append_data_to_data_set(name=data_set_name, data=df, sort=sort)
        logger.info(f'Created data set with id {data_set["id"]} and name {data_set_name}')
        for (col, map_val) in zip(columns, mapping):
            data_mapping_to_tuple[col] = (map_val, data_set, res_sort)
        return data_mapping_to_tuple

    async def _create_dumped_data_set(self, data_set_name: str, data: list[dict]) -> dict:
        """ Get the mapping from a dict.
        :param data: the data
        :return: the mapping tuple
        """
        mapping, data_set, sort = await self._app.append_data_to_data_set(
            name=data_set_name, data=data, dump_whole=True
        )
        logger.info(f'Created data set with id {data_set["id"]} and name {data_set_name}')
        assert mapping[0] == 'customField1'
        return {'data': ('customField1', data_set, sort)}

    async def _create_data_set(
        self, name: str, data: Union[list[dict], pd.DataFrame, dict, str],
        dump_whole: bool = False
    ) -> dict[str, tuple[Mapping, DataSet, dict]]:
        """ Set a shared data entry.
        :param name: the key of the shared data entry
        :param data: the value of the shared data entry
        :param dump_whole: whether to dump the whole data set
        """
        if self.reuse_data_sets:
            data_set = await self._app.get_data_set(name=name, create_if_not_exists=False)
            if data_set:
                return await self._try_to_reuse_data(data_set, data) if not dump_whole else \
                    {'data': ('customField1', data_set, {})}

        await self._delete_data_set_if_exists(data_set_name=name)

        if dump_whole:
            return await self._create_dumped_data_set(name, data)

        return await self._create_data_set_from_df(name, DataFrame(data))

    async def _set_shared_data_entry(
        self, name: str,
        data: Union[list[dict], pd.DataFrame, dict, str],
        dump_whole: bool = False
    ):
        """
        Set a shared data entry.
        :param name: the key of the shared data entry
        :param data: the value of the shared data entry
        :param dump_whole: whether to dump the whole data set
        """
        if name in self._shared_data_map:
            log_error(logger, f'Cannot set shared data entry {name} because it already exists', DataError)

        self._shared_data_map[name] = await self._create_data_set(name, data, dump_whole)
        self._shared_data[name] = data

    async def set_shared_data(
        self, dfs: dict[str, Union[list[dict], pd.DataFrame]] = None, custom_data: dict[str, any] = None
    ):
        """
        Set shared data for the s.plt to use.
        :param dfs: the data frames
        :param custom_data: the custom data
        """
        tasks = []
        for k, df in (dfs if dfs else {}).items():
            tasks.append(self._set_shared_data_entry(k, df))
        for k, v in (custom_data if custom_data else {}).items():
            tasks.append(self._set_shared_data_entry(k, v, dump_whole=True))
        if tasks:
            await asyncio.gather(*tasks)
        else:
            log_error(logger, 'No data provided', DataError)

    async def _choose_data(
            self, order: int, data: Union[list[dict], pd.DataFrame, dict, str],
            chart_class: type[Report] = EChart, dump_whole: bool = False
    ) -> dict[str, tuple[Mapping, DataSet, dict]]:
        """ Get the data mappings of the data. If the data is a string, it is assumed to be a shared data entry.
        :param order: the order of the chart
        :param data: the data
        :param dump_whole: whether to dump the whole data set
        :return: the data mappings
        """
        if isinstance(data, str):
            if data not in self._shared_data_map:
                log_error(logger, f'No shared data with name {data} found', DataError)
            return self._shared_data_map[data]

        _, report = await self._get_chart_report(order, chart_class)
        name = report['id']

        return await self._create_data_set(name, data, dump_whole)

    async def _check_previous_paths(self):
        """ Check if there are paths that have not been executed yet. """
        paths_in_order = await self._app.get_paths_in_order()
        not_executed_paths = [p for p in paths_in_order if p not in self._execution_path_orders]
        if len(not_executed_paths) > 0:
            not_uploaded_paths = [p for p in self._execution_path_orders if p not in paths_in_order]
            self._execution_path_orders.clear()
            for p in paths_in_order + not_uploaded_paths:
                self._execution_path_orders.append(p)

    async def _create_chart(self, chart_class: type[Report], order: int, **params) -> Report:
        """ Create a chart and add it to the current tabs group or modal.
        :param chart_class: the class of the chart
        :param order: the order of the chart in the dashboard
        :param params: the arguments of the chart
        """
        if 'logging_func_name' in params:
            del params['logging_func_name']

        if not params.get('title'):
            params['title'] = ''
        if not params.get('sizeRows'):
            params['sizeRows'] = 3
        if not params.get('sizeColumns'):
            params['sizeColumns'] = 12
        if not params.get('sizePadding'):
            params['sizePadding'] = '0,0,0,0'

        params['bentobox'] = self._get_bentobox_data(order=order)
        params['path'] = self._current_path

        if hasattr(self, 'code_generation_metadata'):
            if 'properties' not in params:
                params['properties'] = {}
            params['properties']['code_generation_metadata'] = getattr(self, 'code_generation_metadata')

        event_type: EventType = EventType.REPORT_UPDATED
        r_hash, chart = await self._get_chart_report(order, chart_class, create_if_not_exists=False)
        if chart:
            params['pathOrder'] = chart['pathOrder']
            if self._current_path and params['pathOrder'] is None:
                await self._check_previous_paths()
                params['pathOrder'] = self._execution_path_orders.index(self._current_path)
            if chart != self._app.mock_create_report(chart_class, r_hash=r_hash, order=order, **params):
                await self._app.update_report(r_hash=r_hash, **params)
                logger.info(f'Updated {chart_class.__name__} at {str(chart)}')
            else:
                logger.info(f'No changes needed for {chart_class.__name__} at {str(chart)}')
                return chart
        else:
            event_type: EventType = EventType.REPORT_CREATED
            await self._check_previous_paths()
            params['pathOrder'] = self._execution_path_orders.index(self._current_path) if self._current_path else None
            params['order'] = order
            chart: Report = await self._app.create_report(chart_class, r_hash=r_hash, **params)
            logger.info(f'Created {chart_class.__name__} at {str(chart)} with id {chart["id"]}')

        added_to_container = False
        if self._current_tabs_group:
            if not self._current_tabs_group.has_report(chart):
                self._current_tabs_group.add_report(tab=self._current_tab, report=chart)
                logger.info(f'Included chart {str(chart)} in tabs group {str(self._current_tabs_group)}')
                added_to_container = True
        elif self._current_modal and not self._current_modal.has_report(chart):
            self._current_modal.add_report(chart)
            logger.info(f'Included chart {str(chart)} in modal {str(self._current_modal)}')
            added_to_container = True

        if added_to_container and 'update_containers' not in self._async_pool.ending_tasks:
            self._async_pool.ending_tasks['update_containers'] = self._update_containers()

        await self._create_event(event_type, {}, chart['id'])

        return chart

    def _get_field_mapping(self, field: Union[str, tuple, list, dict], data_mapping_to_tuple: dict
                           ) -> Union[str, list, dict]:
        """ Get the mapping of a field. """
        if isinstance(field, str):
            return data_mapping_to_tuple[field][0]
        elif isinstance(field, (tuple, list)):
            return [data_mapping_to_tuple[f][0] for f in field]
        elif isinstance(field, dict):
            return {k: data_mapping_to_tuple[v][0] for k, v in field.items()}

    def _check_mapping_in_report_data_set(self, report_data_set: Report.ReportDataSet,
                                          mapping: Mapping) -> bool:
        """ Check if the mapping is correct in the report data set.
        :param report_data_set: the report data set
        :param mapping: the mapping
        :return: True if the mapping is consistent, False otherwise
        """
        report_data_set_mapping = report_data_set['properties']['mapping']
        if report_data_set_mapping != mapping:
            logger.warning(f'Mapping inconsistent in component data set link {str(report_data_set)}')
        return report_data_set_mapping == mapping

    @staticmethod
    def _get_filter_properties(
            series_name: str, converted_data: pd.DataFrame, multi_select: bool
    ) -> tuple[Optional[list[str]], Optional[list[str]], FilterDataSet.InputType]:
        if series_name.startswith('date'):
            return None, ['contains'], FilterDataSet.InputType.DATERANGE
        elif series_name.startswith('int'):
            return None, None, FilterDataSet.InputType.NUMERIC
        elif series_name.startswith('string'):
            return (converted_data[series_name].unique().tolist(), ['eq'],
                    FilterDataSet.InputType.CATEGORICAL_SINGLE
                    if not multi_select else FilterDataSet.InputType.CATEGORICAL_MULTI)
        log_error(logger, f"Field type {series_name} is not supported", NotImplementedError)

    @add_general_async_group
    async def filter(
            self, order: int, data: str, field: str, multi_select: bool = False,
            cols_size: int = 4, rows_size: int = 1, padding: Optional[str] = None
    ):
        """
        Filter the data set.
        :param order: the order of the chart
        :param data: the data
        :param field: the field to filter
        :param multi_select: whether to allow multi select
        :param cols_size: the size of the columns
        :param rows_size: the size of the rows
        :param padding: the padding
        """

        if not isinstance(data, str):
            log_error(logger, f'To filter a data set, the data must be a shared data entry', DataError)

        if data not in self._shared_data_map:
            log_error(logger, f'No shared data with name {data} found', DataError)

        _, data_set, _ = self._shared_data_map[data][field]
        data = self._shared_data[data]

        converted_data, series_name = convert_data_and_get_series_name(data, field)
        options, operations, input_type = self._get_filter_properties(series_name, converted_data, multi_select)

        properties = {
            'filter': [{
                'field': field,
                'inputType': input_type.value,
            }],
            'mapping': [{
                field: series_name,
                'id': data_set['id']
            }]
        }

        if operations:
            properties['filter'][0]['operations'] = operations
        if options:
            properties['filter'][0]['options'] = options

        await self._create_chart(
            chart_class=FilterDataSet,
            order=order,
            title='',
            sizeRows=rows_size,
            sizeColumns=cols_size,
            sizePadding=padding,
            properties=properties,
        )

    @add_general_async_group
    async def iframe(
            self, url: str, order: int, height: int = 640, cols_size: Optional[int] = None,
            padding: Optional[str] = None
    ):
        """ Create an iframe report in the dashboard.
        :param url: the url of the iframe
        :param order: the order of the iframe
        :param height: the height of the iframe
        :param cols_size: the columns that the iframe occupies
        :param padding: padding
        """
        await self._create_chart(
            chart_class=IFrame,
            order=order,
            dataFields=dict(
                url=url,
                height=height,
            ),
            sizePadding=padding,
            sizeColumns=cols_size,
            sizeRows=ceil(height / 240),
        )


    async def _html(
        self, html: str, order: int, cols_size: Optional[int] = None,
        rows_size: Optional[int] = None, padding: Optional[str] = None
    ):
        """ html for internal use """
        await self._create_chart(
            chart_class=HTML,
            order=order,
            chartData=[{'value': html}],
            sizePadding=padding,
            sizeColumns=cols_size,
            sizeRows=rows_size,
        )

    @add_general_async_group
    async def html(
            self, html: str, order: int, cols_size: Optional[int] = None,
            rows_size: Optional[int] = None, padding: Optional[str] = None
    ):
        """
        Create an html report in the dashboard.
        :param html: the html code
        :param order: the order of the html
        :cols_size: the columns that the html occupies
        :rows_size: the rows that the html occupies
        :padding: padding
        """
        await self._html(html, order, cols_size, rows_size, padding)

    @add_general_async_group
    async def single_indicator(
        self, data: dict, order: int, cols_size: Optional[int] = None,
        rows_size: Optional[int] = None, padding: Optional[str] = None,
    ):
        """
        Create an indicator report in the dashboard.
        :param data: the data of the indicator
        :param order: the order of the indicator
        :cols_size: the columns that the indicator occupies
        :rows_size: the rows that the indicator occupies
        :padding: padding
        """
        await self._create_chart(
            chart_class=Indicator,
            path=self._current_path,
            order=order,
            sizePadding=padding,
            sizeRows=rows_size,
            sizeColumns=cols_size,
            properties=data,
        )

    def indicator(
            self, data: Union[str, pd.DataFrame, list[dict], dict], order: int,
            vertical: Union[bool, str] = False, color_by_value: bool = False,
            cols_size: int = 12, rows_size: int = 1, padding: Optional[str] = None,
            **kwargs  # makes backwards compatibility easier
    ):
        """
        Create an indicator report in the dashboard.
        :param data: the data of the indicator
        :param order: the order of the indicator
        :param vertical: whether the indicator is vertical
        :param color_by_value: whether to color the indicator by value
        :cols_size: the columns that the indicator occupies
        :rows_size: the rows that the indicator occupies
        :padding: padding
        :title: the title of the indicator
        """
        df = pd.DataFrame(data if isinstance(data, list) else [data])
        keep_columns = [k for k in df.columns if k in list(Indicator.default_properties.keys())]
        df = df[keep_columns]

        original_padding = padding

        len_df = len(df)
        if vertical and (len_df > 1 or isinstance(vertical, str)):
            if self._bentobox_data:
                log_error(logger, 'Cannot create vertical indicators in a bentobox', RuntimeError)

            self.set_bentobox(cols_size=cols_size, rows_size=rows_size * len_df)

            # fixexd cols_size for bentobox and variable rows size
            cols_size = 22
            rows_size = rows_size * 10 - 2

            padding = '1,1,0,1'
            if isinstance(vertical, str):
                html = (
                    f"<h5 style='font-family: Rubik'>{vertical}</h5>"
                )
                self.html(html=html, order=order, rows_size=2, cols_size=cols_size, padding=padding)
                order += 1
        else:
            bentobox_data = self._bentobox_data
            if padding is None:
                padding = '0,0,0,0'

            padding = padding.replace(' ', '')

            remaining_cols = cols_size % len_df

            extra_padding = remaining_cols // 2
            padding_left_int = int(padding[6]) + extra_padding
            padding_right_int = int(padding[2]) + extra_padding + remaining_cols % 2

            padding_left = f'{padding[0]},0,{padding[4]},{padding_left_int}'
            padding_right = f'{padding[0]},{padding_right_int},{padding[4]},{int(bool(bentobox_data))}'
            padding_else = f'{padding[0]},0,{padding[4]},{int(bool(bentobox_data))}'

            cols_size = cols_size // len_df - int(bool(bentobox_data))
            if cols_size < 2:
                log_error(logger, f'The calculation of the individual cols_size for each indicator '
                                  f'is too small (cols_size/len(df)): {cols_size}', ValueError)

        last_index = df.index[-1]
        first_index = df.index[0]

        for index, df_row in df.iterrows():

            if not vertical:
                if index == first_index and index == last_index:
                    padding = original_padding
                elif index == first_index:
                    padding = padding_left
                elif index == last_index:
                    padding = padding_right
                else:
                    padding = padding_else
            elif index == last_index and vertical and (len_df > 1 or isinstance(vertical, str)):
                padding = '1,1,1,1'

            if color_by_value and 'color' not in df_row and 'icon' not in df_row:
                df_row['color'] = 'success' \
                    if df_row['value'] > 0 else 'error' if df_row['value'] < 0 else 'neutral'
                df_row['icon'] = 'Line/arrow-up' \
                    if df_row['value'] > 0 else 'Line/arrow-down' if df_row['value'] < 0 else 'none'

            self.single_indicator(
                data=df_row.dropna().to_dict(), order=order, cols_size=cols_size,
                rows_size=rows_size, padding=padding
            )

            if isinstance(order, int):
                order += 1

        if vertical:
            self.pop_out_of_bentobox()

        return order

    async def _button(
            self, label: str, order: int,
            rows_size: Optional[int] = None, cols_size: Optional[int] = None,
            align: Optional[str] = 'stretch', padding: Optional[str] = None,
            on_click_events: Optional[Union[list[dict], dict]] = None,
    ):
        """ Create a button in the dashboard. """
        if not on_click_events:
            on_click_events = []
        elif isinstance(on_click_events, dict):
            on_click_events = [on_click_events]

        await self._create_chart(
            chart_class=Button,
            order=order,
            sizePadding=padding,
            sizeRows=rows_size,
            sizeColumns=cols_size,
            properties=dict(
                text=label,
                align=align,
                events=dict(
                    onClick=on_click_events
                )
            )
        )

    @add_general_async_group
    async def button(
            self, label: str, order: int,
            rows_size: Optional[int] = 1, cols_size: int = 2,
            align: Optional[str] = 'stretch', padding: Optional[str] = None,
            on_click_events: Optional[Union[list[dict], dict]] = None,
    ):
        """ Create a button in the dashboard. """
        await self._button(
            label=label, order=order, rows_size=rows_size, cols_size=cols_size,
            align=align, padding=padding, on_click_events=on_click_events
        )

    @add_general_async_group
    async def modal_button(
            self, label: str, order: int, modal: str,
            rows_size: Optional[int] = 1, cols_size: int = 2,
            align: Optional[str] = 'stretch', padding: Optional[str] = None,
    ):
        """
        Create a button in the dashboard that opens a modal.
        :param label: the label of the button
        :param order: the order of the button
        :param modal: the name of the modal
        :param rows_size: the size of the rows in the button
        :param cols_size: the size of the columns in the button
        :param align: the alignment of the button
        :param padding: the padding of the button
        """
        modal_id = (await self._get_modal(modal))['id']

        await self._button(
            label=label, order=order, padding=padding,
            rows_size=rows_size, cols_size=cols_size, align=align,
            on_click_events=[dict(
                action="openModal",
                params=dict(
                    modalId=modal_id,
                )
            )]
        )

    @add_general_async_group
    async def activity_button(
            self, label: str, order: int,
            rows_size: Optional[int] = 1, cols_size: int = 2,
            align: Optional[str] = 'stretch', padding: Optional[str] = None,
            activity_id: Optional[str] = None, activity_name: Optional[str] = None,
    ):
        """
        Create a button in the dashboard that executes an activity.
        :param label: the label of the button
        :param order: the order of the button
        :param rows_size: the size of the rows in the button
        :param cols_size: the size of the columns in the button
        :param align: the alignment of the button
        :param padding: the padding of the button
        :param activity_id: the id of the activity
        :param activity_name: the name of the activity
        """
        activity_id = (await self._app.get_activity(activity_id, activity_name))['id']

        await self._button(
            label=label, order=order, padding=padding,
            rows_size=rows_size, cols_size=cols_size, align=align,
            on_click_events=[dict(
                action="runActivity",
                params=dict(
                    activityId=activity_id,
                )
            )]
        )

    @add_general_async_group
    async def table(
            self, order: int, data: Union[str, pd.DataFrame, list[dict], dict],
            columns: Optional[list[str]] = None, columns_button: bool = True,
            filters: bool = True, export_to_csv: bool = True, search: bool = True,
            page_size: int = 10, page_size_options: Optional[list[int]] = None,
            initial_sort_column: Optional[str] = None, sort_descending: bool = False,
            columns_options: Optional[dict] = None, categorical_columns: Optional[list[str]] = None,
            label_columns: Optional[dict] = None, web_link_column: Optional[str] = None,
            open_link_in_new_tab: bool = False, title: Optional[str] = None,
            padding: Optional[str] = None, rows_size: Optional[int] = None,
            cols_size: Optional[int] = None
    ):
        """ Create a table report in the dashboard.
        :param order: the order of the table
        :param data: the data of the table
        :param columns: the columns of the table
        :param columns_button: whether to show the columns button
        :param filters: whether to show the filters button
        :param export_to_csv: whether to show the export to csv button
        :param search: whether to show the search bar
        :param page_size: the number of rows per page
        :param page_size_options: the options for the number of rows per page
        :param initial_sort_column: the initial sorting column
        :param sort_descending: whether to sort descending by the initial sorting column
        :param columns_options: the options for the columns
        :param categorical_columns: the categorical columns
        :param label_columns: the label columns
        :param report_params: additional report parameters as key-value pairs
        :param web_link_column: the column to use as web link
        :param open_link_in_new_tab: whether to open the web link in a new tab
        :param title: the title of the table
        :param padding: the padding of the table
        :param rows_size: the rows size of the table
        :param cols_size: the columns size of the table
        """

        data_mappings_to_tuples = await self._choose_data(order, data, Table)
        if isinstance(data, str):
            data = self._shared_data[data]

        df = validate_data_is_pandarable(data)

        if not columns:
            columns = list(data_mappings_to_tuples.keys())
        if not columns_options:
            columns_options = {}
        if not categorical_columns:
            categorical_columns = []
        if not label_columns:
            label_columns = {}
        else:
            aux_label_columns = label_columns
            label_columns = {}
            for has_to_be_tuple in aux_label_columns:
                v = aux_label_columns[has_to_be_tuple]
                if isinstance(has_to_be_tuple, str):
                    label_columns[(has_to_be_tuple, 'filled')] = v
                elif isinstance(has_to_be_tuple, tuple):
                    label_columns[has_to_be_tuple] = v
                else:
                    log_error(logger, f'Invalid label_columns key: {has_to_be_tuple}', ValueError)
        label_just_columns = [x[0] for x in label_columns.keys()]

        _, data_set, _ = data_mappings_to_tuples[columns[0]]
        columns_dicts = []
        rows_dict = {'mapping': {}}
        for i, name in enumerate(columns):
            column_options = {'field': name, 'headerName': name, 'order': i}

            if name in columns_options:
                column_options.update(columns_options[name])

            if name in categorical_columns:
                if name in label_columns:
                    log_error(logger,
                              f'Columns cannot be both included in categorical_columns and label_columns: {name}',
                              ValueError)
                column_options['type'] = 'singleSelect'
                column_options['options'] = df[name].unique().tolist()

            if name in label_just_columns:
                index = label_just_columns.index(name)
                (_, variant), label_options = list(label_columns.items())[index]
                column_options['chips'] = {}
                column_options['chips']['variant'] = variant
                column_options['chips']['options'] = interpret_label_info(df, name, label_options, variant)

            if name == web_link_column:
                column_options['link'] = {'url': 'webLink', 'openNewTab': open_link_in_new_tab}

            columns_dicts.append(column_options)
            rows_dict['mapping'][name] = data_mappings_to_tuples[name][0]

        if web_link_column:
            rows_dict['mapping']['web'] = data_mappings_to_tuples[web_link_column][0]
            rows_dict['mapping']['webLink'] = data_mappings_to_tuples[web_link_column][0]

        _, report = await self._get_chart_report(order, Table)

        rds = await report.get_report_data_sets()

        if self.reuse_data_sets and report['properties'].get('columns') and \
                report['properties']['columns'] != columns_dicts:
            log_error(logger, f"Columns options do not match the previous columns, most likely data has changed,"
                              f" don't use the reuse_data_sets option in this case", DataError)

        if rds and (not self.reuse_data_sets or any(rd['dataSetId'] != data_set['id'] for rd in rds)):
            await report.delete_report_data_sets(log=True)
            rds = []

        if not rds:
            await report.create_report_data_set((None, data_set, None))

        pagination = {'pageSize': page_size}
        if page_size_options:
            pagination['pageSizeOptions'] = page_size_options

        sort = {}
        if initial_sort_column:
            sort['field'] = data_mappings_to_tuples[initial_sort_column][0]
            sort['direction'] = 'asc' if not sort_descending else 'desc'

        await self._create_chart(
            chart_class=Table,
            order=order,
            sizePadding=padding,
            sizeRows=rows_size,
            sizeColumns=cols_size,
            title=title,
            properties=dict(
                columns=columns_dicts,
                rows=rows_dict,
                columnsButton=columns_button,
                filtersButton=filters,
                exportButton=export_to_csv,
                search=search,
                pagination=pagination,
                sort=sort,
            )
        )

    @add_general_async_group
    async def input_form(
            self, options: dict, order: int, padding: Optional[str] = None,
            rows_size: Optional[int] = None, cols_size: Optional[int] = None,
            modal: Optional[str] = None, activity_id: Optional[str] = None,
            activity_name: Optional[str] = None, on_submit_events: Optional[list[dict]] = None
    ):
        """ Creates an input form.
        :param options: the options for the input form
        :param order: the order of the input form
        :param padding: the padding of the input form
        :param rows_size: the rows size of the input form
        :param cols_size: the columns size of the input form
        :param modal: the modal to open after submitting the form
        :param activity_id: the activity id to run after submitting the form
        :param activity_name: the activity name to run after submitting the form
        :param on_submit_events: the events to run after submitting the form
        """
        validate_input_form_data(self, options)

        if on_submit_events is None:
            on_submit_events = []

        if modal:
            modal_id = (await self._get_modal(modal))['id']
            on_submit_events.append({
                'action': 'openModal',
                'params': {
                    'modalId': modal_id
                }
            })

        if activity_id or activity_name:
            activity_id = (await self._app.get_activity(activity_id, activity_name))['id']

            on_submit_events.append({
                "action": "runActivity",
                "params": {
                    "activityId": activity_id,
                }
            })

        initial_data: dict = {}
        for fields in options['fields']:
            for field in fields['fields']:
                field_name: str = field['fieldName']
                input_type: Optional[str] = field.get('inputType')
                if input_type == 'color':
                    initial_data[field_name] = '#000000'
                elif input_type == 'number':
                    initial_data[field_name] = 0
                else:
                    initial_data[field_name] = ''

        r_hash, report = await self._get_chart_report(order, InputForm)
        _, data_set, _ = list((await self._create_data_set(report['id'], initial_data, dump_whole=True)).values())[0]

        rds = await report.get_report_data_sets()
        if len(rds) > 1:
            log_error(logger, f"Input form can only have one component data set link, "
                              f"this is an SDK bug please report it and "
                              f"delete the form to solve the issue", DataError)

        if rds and (not self.reuse_data_sets or rds[0]['dataSetId'] != data_set['id']
                    or rds[0]['properties']['fields'] != options['fields']):
            await report.delete_report_data_sets(log=True)
            rds = []

        if not rds:
            await report.create_report_data_set(('customField1', data_set, None), properties=options)
            logger.info(f"Created 1 component data set link for input form at {r_hash}")

        await self._create_chart(
            chart_class=InputForm,
            order=order,
            sizePadding=padding,
            sizeRows=rows_size,
            sizeColumns=cols_size,
            properties={'events': {'onSubmit': on_submit_events}}
        )

    def generate_input_form_groups(
            self, order: int, form_groups: dict,
            dynamic_sequential_show: Optional[bool] = False,
            auto_send: Optional[bool] = False,
            next_group_label: Optional[str] = 'Next',
            rows_size: Optional[int] = 3, cols_size: int = 12,
            padding: Optional[str] = None,
            modal: Optional[str] = None,
            activity_id: Optional[str] = None,
            activity_name: Optional[str] = None,
            on_submit_events: Optional[list[dict]] = None,
    ):
        """ Easier way to create an input form.
        :param menu_path: the menu path of the input form
        :param order: the order of the input form
        :param form_groups: the form groups of the input form
        :param dynamic_sequential_show: whether to show the next group after submitting the current group
        :param auto_send: whether to automatically send the form after the last group
        :param next_group_label: the label of the next group button
        :param rows_size: the rows size of the input form
        :param cols_size: the columns size of the input form
        :param padding: the padding of the input form
        :param modal: the modal to open after submitting the form
        :param activity_id: the activity id to run after submitting the form
        :param activity_name: the activity name to run after submitting the form
        :param on_submit_events: the events to run after submitting the form
        """
        report_data_set_properties = {'fields': []}
        if auto_send:
            report_data_set_properties['variant'] = 'autoSend'

        r_hash = self._get_chart_hash(order)

        next_id = f'{r_hash}_0' if dynamic_sequential_show else None

        next_group_index = 1
        for form_group_name, form_group in form_groups.items():
            form_group_json = {'title': form_group_name, 'fields': form_group}
            if next_id:
                form_group_json['id'] = next_id
                next_id = f'{r_hash}_{next_group_index}'
                form_group_json['nextFormGroup'] = {'id': next_id, 'label': next_group_label}
                next_group_index += 1
            report_data_set_properties['fields'] += [form_group_json]

        if next_id:
            del report_data_set_properties['fields'][-1]['nextFormGroup']

        self.input_form(
            options=report_data_set_properties,
            order=order, rows_size=rows_size, cols_size=cols_size,
            padding=padding, modal=modal, activity_id=activity_id, activity_name=activity_name,
            on_submit_events=on_submit_events,
        )

    async def _get_input_form_data(self, input_form: InputForm) -> dict:
        """ Get the input form data. """
        report_data_set = (await input_form.get_report_data_sets())[0]
        data_set = await self._app.get_data_set(report_data_set['dataSetId'])
        data_point = await data_set.get_one_data_point()
        logger.info(f"Got input form data for {str(input_form)}")
        return {
            'reportId': input_form['id'],
            'order': input_form['order'],
            'data': data_point['customField1'] if data_point else None,
        }

    async def get_input_forms(self) -> list[dict]:
        """ Get all the input forms in the current menu_path. """
        reports: list = [report for report in await self._app.get_reports()
                         if report['reportType'] == 'FORM'
                         and (report['path'] == self._current_path or not self._current_path)]
        return list(await asyncio.gather(*[self._get_input_form_data(report) for report in reports]))

    @add_general_async_group
    async def annotated_chart(
            self, data: Union[list[DataFrame], list[list[dict]]], order: int,
            x: str, y: Union[str, list[str]], annotations: str = 'annotation',
            rows_size: Optional[int] = None, cols_size: Optional[int] = None,
            padding: Optional[str] = None,
            title: Optional[str] = None,
            y_axis_name: Optional[str] = None,
            slider_config: Optional[dict] = None,
            slider_marks: Optional[list[tuple]] = None,
    ):
        """
        Create an annotated chart in the menu path.
        :param data: the data of the chart
        :param order: the order of the chart
        :param x: the x axis
        :param y: the y axis
        :param annotations: the annotations
        :param rows_size: the rows size of the chart
        :param cols_size: the columns size of the chart
        :param padding: the padding of the chart
        :param title: the title of the chart
        :param y_axis_name: the name of the y axis
        :param slider_config: the slider config
        :param slider_marks: the slider marks
        """
        if isinstance(data, str):
            log_error(logger, f'Annotated chart does not support the use of shared data', DataError)

        if slider_config is None:
            slider_config = {}

        if slider_marks:
            slider_config['marks'] = [{'label': mark[0], 'value': mark[1]} for mark in slider_marks]

        if isinstance(y, str):
            y = [y]

        if not len(y) == len(data):
            log_error(logger, f'Number of y values ({len(y)}) does not match number of dataframes ({len(data)})',
                      DataError)

        _, report = await self._get_chart_report(order, AnnotatedEChart)

        dfs = [validate_data_is_pandarable(df) for df in data]
        data_set_tasks = []
        for df, y_val in zip(dfs, y):
            df[x] = pd.to_datetime(df[x])
            if annotations in df:
                df[annotations] = df[annotations].replace(np.nan, '', regex=True).astype(str)
            data_set_tasks.append(self._create_data_set(name=f"{report['id']}_{y_val}", data=df))

        data_set_dicts = await asyncio.gather(*data_set_tasks)
        data_sets = [ds[y_val][1] for ds, y_val in zip(data_set_dicts, y)]

        rd_ids = None
        if self.reuse_data_sets:
            rd_ids = get_uuids_from_dict(report['properties']['option'])
            if len(rd_ids) == len(y):
                rep_ds: list[Report.ReportDataSet] = await report.get_report_data_sets()
                rep_ds = [next((rd for rd in rep_ds if rd['id'] == rd_id), None) for rd_id in rd_ids]
                if any(rd['dataSetId'] != ds['id'] or
                       rd['properties']['mapping'] != {'values': ['dateField1', 'intField1'], 'label': 'stringField1'}
                       for rd, ds in zip(rep_ds, data_sets)):
                    rd_ids = None

        if not rd_ids:
            await report.delete_report_data_sets(log=True)
            mapping: Mapping = {'values': ['dateField1', 'intField1'], 'label': 'stringField1'}
            report_data_sets = await asyncio.gather(*[
                report.create_report_data_set((mapping, data_set, None)) for data_set in data_sets
            ])
            rd_ids = [rd['id'] for rd in report_data_sets]

        chart_options = {
            'yAxis': {
                'name': y_axis_name or '',
                'font'
                'type': 'value',
                'nameLocation': 'middle',
                'nameGap': 30,
            },
            'xAxis': {
                'type': 'time',
                'nameLocation': 'middle',
                'nameGap': 30,
            },
            'series': [
                {
                    'data': '#{' + rd_id + '}',
                    'type': 'line',
                    'name': y_name,
                    'itemStyle': {
                        'borderRadius': [9, 9, 0, 0],
                    }
                }
                for y_name, rd_id in zip(y, rd_ids)
            ],
        }
        await self._create_chart(
            chart_class=AnnotatedEChart,
            order=order,
            sizePadding=padding,
            sizeRows=rows_size,
            sizeColumns=cols_size,
            title=title,
            properties=dict(
                slider=slider_config,
                option=chart_options,
            )
        )

    async def _get_report_data_sets_per_mapping(
            self, report: Report, data_mapping_to_tuple: dict, fields: list[Union[tuple, dict]],
    ) -> list[Report.ReportDataSet]:
        """ Create a data_set per field of a dataframe.
        :param report: the report
        :param data_mapping_to_tuple: the mapping of the data
        :param fields: the fields of the data to be used
        :return: the list of data sets partitioned by axis and fields
        """
        tasks = []
        fields = fields if fields else []
        for i, f in enumerate(fields):
            mapping = []
            data_set = None
            res_sort = None
            if isinstance(f, str):
                if f not in data_mapping_to_tuple:
                    log_error(logger, f'Field {f} not found in data', DataError)
                mapping, data_set, res_sort = data_mapping_to_tuple[f]
            elif isinstance(f, (tuple, list)):
                for f_ in f:
                    if f_ not in data_mapping_to_tuple:
                        log_error(logger, f'Field {f_} not found in data', DataError)
                    mapping_, data_set_, res_sort_ = data_mapping_to_tuple[f_]
                    mapping.append(mapping_)
                    data_set = data_set if data_set else data_set_
                    res_sort = res_sort if res_sort else res_sort_
                    assert data_set == data_set_
                    assert res_sort == res_sort_
            elif isinstance(f, dict):
                mapping = {k: [] for k in f.keys()}
                for k, f_ in f.items():
                    if f_ not in data_mapping_to_tuple:
                        log_error(logger, f'Field {f_} not found in data', DataError)
                    mapping_, data_set_, res_sort_ = data_mapping_to_tuple[f_]
                    mapping[k] = mapping_
                    data_set = data_set if data_set else data_set_
                    res_sort = res_sort if res_sort else res_sort_
                    assert data_set == data_set_
                    assert res_sort == res_sort_
            else:
                log_error(logger, f'Field {f} is not a string, tuple, list or dict', DataError)
            tasks.append(report.create_report_data_set(mapping_data_set_sort=(mapping, data_set, res_sort)))

        report_data_sets = await asyncio.gather(*tasks)
        logger.info(f'Created {len(report_data_sets)} component data set links for component {str(report)}')
        return report_data_sets

    def _update_options(self, options: dict, option_modifications: Optional[dict]):
        """ Update the options of an echart.
        :param options: the options of the echart
        :param option_modifications: the modifications to apply to the options
        """
        if not option_modifications:
            return
        for k, v in option_modifications.items():
            if k == 'series':
                log_error(logger, 'Series cannot be modified', DataError)
            if k == 'xAxis' or k == 'yAxis':
                if isinstance(v, dict):
                    v = [v]
                if len(v) != len(options[k]):
                    log_error(logger, f'The number of {k} must be {len(v)}', DataError)
                for i, v_ in enumerate(v):
                    options[k][i].update(v_)
            else:
                options[k] = v

    async def _create_echart(
            self, options: dict, order: int,
            fields: list[Union[str, tuple, dict]],
            data_mapping_to_tuples: Optional[dict] = None,
            data: Optional[Union[str, pd.DataFrame]] = None,
            dump_whole: bool = False,
            option_modifications: Optional[dict] = None,
            title: Optional[str] = None,
            rows_size: Optional[int] = None,
            cols_size: Optional[int] = None,
            padding: Optional[str] = None,
    ):
        """
        Create an echart in the dashboard, fill in the data referenced in the echart_options and create or
        update the chart and data set links.
        :param options: the options of the echart
        :param data_mapping_to_tuples: the mapping of the data to the tuples
        :param data: the data of the echart
        :param dump_whole: whether to dump the whole data
        :param option_modifications: the modifications to apply to the options
        :param title: the title of the echart
        :param rows_size: the rows size of the echart
        :param cols_size: the columns size of the echart
        :param padding: the padding of the echart
        """

        if data is not None:
            data_mapping_to_tuples = await self._choose_data(order, data, dump_whole=dump_whole)

        data_key_entries = get_data_references_from_dict(options)
        _, report = await self._get_chart_report(order, EChart)

        rd_ids = None
        if self.reuse_data_sets:
            rd_ids = get_uuids_from_dict(report['properties']['option'])
            if len(rd_ids) == len(fields):
                rep_ds: list[Report.ReportDataSet] = await report.get_report_data_sets()
                mapped_f = [self._get_field_mapping(field, data_mapping_to_tuples) for field in fields]
                for i, rd_id in enumerate(rd_ids):
                    report_data_set = next((rd for rd in rep_ds if rd['id'] == rd_id), None)
                    if not report_data_set or not self._check_mapping_in_report_data_set(report_data_set, mapped_f[i]):
                        rd_ids = None
                        break
            else:
                rd_ids = None

        if not rd_ids:
            await report.delete_report_data_sets(log=True)
            report_data_sets = await self._get_report_data_sets_per_mapping(report, data_mapping_to_tuples, fields)
            rd_ids = [rd['id'] for rd in report_data_sets]

        self._update_options(options, option_modifications)

        if len(rd_ids) != len(data_key_entries):
            log_error(logger, f'The number of data references and fields must be equal, '
                              f'they are {len(data_key_entries)} and {len(rd_ids)} respectively.', DataError)

        for i, data_key_entry in enumerate(data_key_entries):
            data = options
            for key in data_key_entry[:-1]:
                data = data[key]
            data[data_key_entry[-1]] = '#{' + rd_ids[i] + '}'

        await self._create_chart(
            chart_class=EChart,
            properties={'option': options},
            order=order,
            title=title,
            sizePadding=padding,
            sizeRows=rows_size,
            sizeColumns=cols_size,
        )

    @staticmethod
    def _apply_variant(echart_options: dict, variant: Optional[str]):
        if variant == 'clean':
            echart_options.update({'toolbox': {'show': False}, 'legend': {'show': False}, 'grid': {}})
            for axislist in [echart_options['xAxis'], echart_options['yAxis']]:
                for axis in axislist:
                    axis.update({'axisLine': {'show': False}, 'axisTick': {'show': False}})
        elif variant == 'minimal':
            echart_options.update({
                'toolbox': {'show': False}, 'legend': {'show': False},
                'grid': {'left': '1%', 'right': '1%', 'top': '1%', 'bottom': '1%'},
                'tooltip': {'show': False},
            })
            for axislist in [echart_options['xAxis'], echart_options['yAxis']]:
                for axis in axislist:
                    axis.update({'axisLine': {'show': False}, 'axisTick': {'show': False},
                                 'splitLine': {'show': False}, 'axisLabel': {'show': False}})

    async def _create_trend_chart(
            self, axes: Union[list[str], str], order: int,
            data_mapping_to_tuples: Optional[dict], echart_options: dict,
            series_options: Union[dict, list[dict]],
            values: Optional[Union[list[Union[str, tuple]], str, tuple]] = None,
            x_axis_names: Optional[Union[str, list[str]]] = None,
            y_axis_names: Optional[Union[str, list[str]]] = None,
            show_values: Optional[Union[list[str], str]] = None,
            variant: Optional[str] = None,
            **report_params
    ):
        """
        Create a line chart in the dashboard.
        :param data: the data of the chart
        :param axes: the names of the columns to be used as axis
        :param values: the names of the columns to be used as values
        :param x_axis_names: the name of the x axes
        :param y_axis_names: the name of the y axes
        :param echart_options: the options of the chart
        :param series_options: the options of the series of the chart
        :param bottom_toolbox: whether to show the toolbox on top of the chart
        :param order: the order of the chart in the dashboard
        :param report_params: additional report parameters as key-value pairs
        """
        self._apply_variant(echart_options, variant)
        if isinstance(axes, str):
            axes = [axes]
        if isinstance(values, str) or isinstance(values, tuple):
            values = [values]
        if show_values is None:
            show_values = []

        if 'x_axis_name' in report_params:
            x_axis_names = [report_params['x_axis_name']]
        elif isinstance(x_axis_names, str):
            x_axis_names = [x_axis_names]
        else:
            x_axis_names = []

        if 'y_axis_name' in report_params:
            y_axis_names = [report_params['y_axis_name']]
        elif isinstance(y_axis_names, str):
            y_axis_names = [y_axis_names]
        else:
            y_axis_names = []

        values = [v for v in data_mapping_to_tuples.keys() if v not in axes] if values is None else values
        if isinstance(series_options, dict):
            series_options = [series_options] * len(values)
        elif len(series_options) < len(values):
            series_options = series_options + [series_options[-1]] * (len(values) - len(series_options))
        elif len(series_options) > len(values):
            series_options = series_options[:len(values)]

        echart_options['series'] = [
            deep_update({
                'name': name if isinstance(name, str) else ' × '.join(name),
                'label': {'show': show_values == 'all' or name in show_values},
            }, series_options[i])
            for i, name in enumerate(values)
        ]

        for i, axis_options in enumerate(echart_options['xAxis']):
            axis_options['name'] = x_axis_names[i] if i < len(x_axis_names) else None
            if axis_options['name'] and 'nameGap' not in axis_options:
                axis_options['nameGap'] = 32

        for i, axis_options in enumerate(echart_options['yAxis']):
            axis_options['name'] = y_axis_names[i] if i < len(y_axis_names) else None
            if axis_options['name'] and 'nameGap' not in axis_options:
                axis_options['nameGap'] = -24
                axis_options['offset'] = 36
                echart_options['grid']['left'] = '4%'
                axis_options['axisTick'] = {'show': False}
                axis_options['axisLine'] = {'show': False}

        await self._create_echart(
            order=order,
            options=echart_options,
            data_mapping_to_tuples=data_mapping_to_tuples,
            fields=axes + values,
            **report_params
        )

    @add_general_async_group
    async def free_echarts(
            self, data: Optional[Union[str, DataFrame, list[dict]]] = None,
            options: Optional[dict] = None,
            raw_options: Optional[str] = None,
            order: Optional[int] = None,
            title: Optional[str] = None,
            rows_size: Optional[int] = None,
            cols_size: Optional[int] = None,
            padding: Optional[str] = None,
            fields: Optional[list] = None,
            data_is_not_df: bool = False,
    ):
        """
        Create an echarts chart with custom options.
        :param data: the data of the chart
        :param options: the options of the chart
        :param raw_options: the raw options of the chart
        :param order: the order of the chart in the dashboard
        :param title: the title of the chart
        :param rows_size: the rows size of the chart
        :param cols_size: the columns size of the chart
        :param padding: the padding of the chart
        :param fields: the fields of the chart
        :param data_is_not_df: whether the data is not a dataframe
        """
        if not options and not raw_options:
            log_error(logger, 'Either options or raw_options must be provided', ValueError)
        if options and not isinstance(data, DataFrame) and data is None:
            log_error(logger, 'data must be provided when options is provided', DataError)
        if options and raw_options:
            log_error(logger, 'Only one of options and raw_options can be provided', ValueError)

        if raw_options:
            options = transform_dict_js_to_py(raw_options)
            data = retrieve_data_from_options(options)
            options['dataset'] = {'source': '#set_data#'}
            fields = [list(data[0].keys())]
        else:
            options = deepcopy(options)

        await self._create_echart(
            order=order, data_mapping_to_tuples=await self._choose_data(order, data, dump_whole=data_is_not_df),
            fields=fields, options=options, title=title,
            rows_size=rows_size, cols_size=cols_size, padding=padding
        )

    # ECharts
    line = add_general_async_group(line_chart)
    bar = add_general_async_group(bar_chart)
    stacked_bar = add_general_async_group(stacked_bar_chart)
    area = add_general_async_group(area_chart)
    stacked_area = add_general_async_group(stacked_area_chart)
    scatter = add_general_async_group(scatter_chart)
    horizontal_bar = add_general_async_group(horizontal_bar_chart)
    stacked_horizontal_bar = add_general_async_group(stacked_horizontal_bar_chart)
    zero_centered_bar = add_general_async_group(zero_centered_bar_chart)
    funnel = add_general_async_group(funnel_chart)
    tree = add_general_async_group(tree_chart)
    radar = add_general_async_group(radar_chart)
    pie = add_general_async_group(pie_chart)
    doughnut = add_general_async_group(doughnut_chart)
    rose = add_general_async_group(rose_chart)
    sunburst = add_general_async_group(sunburst_chart)
    treemap = add_general_async_group(treemap_chart)
    sankey = add_general_async_group(sankey_chart)
    heatmap = add_general_async_group(heatmap_chart)
    predictive_line = add_general_async_group(predictive_line_chart)
    speed_gauge = add_general_async_group(speed_gauge_chart)
    shimoku_gauge = add_general_async_group(shimoku_gauge_chart)
    shimoku_gauges_group = add_general_async_group(shimoku_gauges_group)
    gauge_indicator = add_general_async_group(gauge_indicator)
    top_bottom_area = add_general_async_group(top_bottom_area_charts)
    top_bottom_line = add_general_async_group(top_bottom_line_charts)
    line_with_confidence_area = add_general_async_group(line_with_confidence_area_chart)
    scatter_with_effect = add_general_async_group(scatter_with_effect_chart)
    waterfall = add_general_async_group(waterfall_chart)
    line_and_bar_charts = add_general_async_group(line_and_bar_charts)
    segmented_line = add_general_async_group(segmented_line_chart)
    marked_line = add_general_async_group(marked_line_chart)
    segmented_area = add_general_async_group(segmented_area_chart)

    # Bentobox charts defined in the bentobox_charts.py file
    infographics_text_bubble = add_general_async_group(infographics_text_bubble)
    chart_and_modal_button = add_general_async_group(chart_and_modal_button)
    chart_and_indicators = add_general_async_group(chart_and_indicators)
    indicators_with_header = add_general_async_group(indicators_with_header)
    line_with_summary = add_general_async_group(line_with_summary)



from shimoku.async_execution_pool import AutoAsyncExecutionPool

# TODO fix this file
from shimoku.api.resources.universe import Universe
from shimoku.api.resources.business import Business
from shimoku.api.resources.app import App
from shimoku.api.resources.dashboard import Dashboard

import shimoku_components_catalog.html_components

from typing import Optional, Dict, List, Union
from uuid import uuid4
from shimoku import Client

import logging
from shimoku.execution_logger import configure_logging
logger = logging.getLogger(__name__)


class MockApiClient:
    """ Mock Api Client class """

    semaphore_limit = 10
    call_counter = 0
    cache_enabled = True
    playground = False

    async def query_element(
            self, method: str, endpoint: str, limit: Optional[int] = None, **kwargs
    ) -> Union[Dict, List, bool]:
        """ MOCK. Retrieve an element if the endpoint exists

        :param method: examples are 'GET', 'POST', etc
        :param endpoint: example: 'business/{businessId}/app/{appId}
        :param limit: limit the number of results returned
        """
        body_params = kwargs.get('body_params', {})
        if isinstance(body_params, list) and method == 'POST':
            return True

        uuid = str(uuid4())
        if method == 'GET' or method == 'PATCH':
            uuid = endpoint.split('/')[-1]
            if '-' not in uuid:
                return []

        body_params.update(dict(id=uuid))

        if 'triggerWebhook' in endpoint:
            body_params['STATUS'] = 'OK'

        return body_params


class MockClient(Client):
    """ Mock Client class """

    def __init__(
            self, verbosity: str = None, async_execution: bool = False
    ):
        self.playground = False
        self._access_token = None
        self.access_token = None
        self.environment = None
        self.universe_id = None
        self.workspace_id = None
        self.board_id = None
        self.menu_path_id = None

        self.configure_logging = configure_logging
        if verbosity:
            self.configure_logging(verbosity)

        self._api_client = MockApiClient()

        self._universe_object = Universe(self._api_client, uuid='1')
        self._business_object: Optional[Business] = None
        self._app_object: Optional[App] = None
        self._dashboard_object: Optional[Dashboard] = None

        self.epc = AutoAsyncExecutionPool(api_client=self._api_client)

        if async_execution:
            self.activate_async_execution()
        else:
            self.activate_sequential_execution()

        self.universes = UniversesLayer(self._api_client, self.epc)
        self.workspaces = WorkspacessLayer(self._universe_object, self.epc)
        self.boards = BoardsLayer(self._business_object, self.epc)
        self.menu_paths = AppsLayer(self._business_object, self.epc)
        self.components = ReportMetadataApi(self._app_object, self.epc)
        self.data = DataSetManagingApi(self._app_object, self.epc)
        self.io = FileMetadataApi(self._app_object, self.epc)
        self.activities = ActivityMetadataApi(self._app_object, self.epc)
        self.plt = PlotLayer(self._app_object, self.epc)
        self._reuse_data_sets = False
        self._shared_dfs = {}
        self._shared_custom_data = {}

        self.epc.current_app = self._app_object
        self.epc.universe = self._universe_object

        self.html_components = shimoku_components_catalog.html_components



from copy import deepcopy
from typing import Optional, Union
from abc import ABC, abstractmethod

from shimoku.api.resources.report import Report
from shimoku.plt.utils import ShimokuPalette
from shimoku.utils import get_default_args

import logging
from shimoku.execution_logger import log_error
logger = logging.getLogger(__name__)


class EChart(Report):
    """ ECharts report class """
    report_type = 'ECHARTS2'

    default_properties = dict(
        **Report.default_properties,
        option={},
    )


default_toolbox_options = {
    'show': True,
    'orient': 'horizontal',
    'itemSize': 20,
    'itemGap': 24,
    'showTitle': True,
    'zlevel': 100,
    'bottom': 0,
    'right': '24px',
    'feature': {
        'dataView': {
            'title': 'data',
            'readOnly': False,
            'icon': 'image://https://uploads-ssl.webflow.com/619f9fe98661d321dc3beec7/6398a555461a3684b16d544e_database.svg',
            'emphasis': {
                'iconStyle': {
                    'textBackgroundColor': ShimokuPalette.CHART_C1.value,
                    'textBorderRadius': [50, 50, 50, 50],
                    'textPadding': [8, 8, 8, 8],
                    'textFill': ShimokuPalette.WHITE.value
                }
            },
        },
        'magicType': {
            'type': ['line', 'bar'],
            'title': {
                'line': 'Switch to Line Chart',
                'bar': 'Switch to Bar Chart',
            },
            'icon': {
                'line': 'image://https://uploads-ssl.webflow.com/619f9fe98661d321dc3beec7/6398a55564d52c1ba4d9884d_linechart.svg',
                'bar': 'image://https://uploads-ssl.webflow.com/619f9fe98661d321dc3beec7/6398a5553cc6580f8e0edea4_barchart.svg'
            },
            'emphasis': {
                'iconStyle': {
                    'textBackgroundColor': ShimokuPalette.CHART_C1.value,
                    'textBorderRadius': [50, 50, 50, 50],
                    'textPadding': [8, 8, 8, 8],
                    'textFill': ShimokuPalette.WHITE.value
                }
            },
        },
        'saveAsImage': {
            'show': True,
            'title': 'Save as image',
            'icon': 'image://https://uploads-ssl.webflow.com/619f9fe98661d321dc3beec7/6398a555662e1af339154c64_download.svg',
            'emphasis': {
                'iconStyle': {
                    'textBackgroundColor': ShimokuPalette.CHART_C1.value,
                    'textBorderRadius': [50, 50, 50, 50],
                    'textPadding': [8, 8, 8, 8],
                    'textFill': ShimokuPalette.WHITE.value
                }
            }
        }
    }
}


def get_common_series_options() -> dict[str, any]:
    return deepcopy({
        'data': '#set_data#',
        'emphasis': {'focus': 'series'},
        'smooth': True,
        'itemStyle': {'borderRadius': [9, 9, 0, 0]},
    })


def get_common_echart_options() -> dict[str, any]:
    # toolbox_options = deepcopy(default_toolbox_options)
    # if bottom_toolbox:
    #     del toolbox_options['top']
    #     toolbox_options['bottom'] = "0px"
    return deepcopy({
        'legend': {
            'show': True,
            'type': 'scroll',
            'icon': 'circle',
            'padding': [5, 5, 5, 5],
        },
        'tooltip': {
            'trigger': 'item',
            'axisPointer': {'type': 'cross'},
        },
        'toolbox': default_toolbox_options,
        'xAxis': [{
            'data': '#set_data#',
            'type': 'category',
            'fontFamily': 'Rubik',
            'name': "",
            'nameLocation': 'middle',
        }],
        'yAxis': [{
            'name': "",
            'nameLocation': 'middle',
            'type': 'value',
            'fontFamily': 'Rubik',
        }],
        'grid': {
            'left': '1%',
            'right': '2%',
            'bottom': 48,
            'containLabel': True
        },
        'series': [get_common_series_options()],
    })


generic_echart_function_params = dict(
    fields='fields', options='options', data='data', option_modifications='option_modifications',
    order='order', title='title', rows_size='rows_size', cols_size='cols_size', padding='padding'
)


class InvertibleOperation(ABC):
    @abstractmethod
    def __call__(self, *args, **kwargs):
        pass

    @abstractmethod
    def revert(self, *args, **kwargs):
        pass


class InvertibleDictFunction(ABC):

    def __init__(
        self, echarts_func: callable, dict_var_name: str = 'options',
        diff_dict_var_name: str = 'option_modifications'
    ):
        self.function_operations = []
        self.params = get_default_args(echarts_func)
        self.params.pop('self')
        self.params[dict_var_name] = {}
        self.params[diff_dict_var_name] = {}
        self.d = self.params[dict_var_name]
        self.diff_dict = self.params[diff_dict_var_name]
        self.dict_var_name = dict_var_name
        self.diff_dict_var_name = diff_dict_var_name
        self.aux_definition_params = deepcopy(self.params)

    class SetItems(InvertibleOperation):

        def __init__(
            self, idf: 'InvertibleDictFunction',
            items: dict, offset: Optional[list[Union[str, int]]] = None,
            final: bool = False,
        ):
            self.d = idf.d
            self.items = deepcopy(items)
            self.previous_items = {}
            self.offset = offset if offset else []
            self.final = final
            self.diff_dict = idf.diff_dict
            self(idf.aux_definition_params[idf.dict_var_name])

        def __call__(self, d: Optional[dict] = None):
            in_definition = d is not None
            d = self.d if d is None else d
            for offset_key in self.offset:
                d = d[offset_key]
            if in_definition:
                self.previous_items = deepcopy(d)
            d.update(deepcopy(self.items))

        def revert(self) -> bool:
            d = self.d
            aux_diff = deepcopy(self.diff_dict)
            for offset_key in self.offset:
                if isinstance(d, list):
                    if offset_key not in aux_diff:
                        aux_diff = [{} for _ in range(offset_key + 1)]
                    if not isinstance(offset_key, int) or offset_key >= len(d):
                        return False
                elif offset_key not in d:
                    return False
                elif offset_key not in aux_diff:
                    aux_diff[offset_key] = {}
                d = d[offset_key]
                aux_diff = aux_diff[offset_key]

            exists_diff = False
            for k in d.keys():
                if k not in self.previous_items and k not in self.items:
                    if self.final:
                        return False
                    aux_diff[k] = d[k]
                    exists_diff = True

            for k in self.items.keys():
                if d[k] != self.items[k]:
                    if self.final:
                        return False
                    aux_diff[k] = d[k]
                    exists_diff = True
                if k in self.previous_items:
                    d[k] = self.previous_items[k]
                else:
                    del d[k]
            if exists_diff:
                self.diff_dict.clear()
                self.diff_dict.update(aux_diff)
            return True

    def set_items(
        self, items: dict, offset: Optional[list[Union[str, int]]] = None, final: bool = False
    ):
        self.function_operations.append(self.SetItems(self, items, offset, final))

    class DeleteItems(InvertibleOperation):
        def __init__(
            self, idf: 'InvertibleDictFunction',
            keys: list[str], offset: Optional[list[Union[str, int]]] = None
        ):
            self.d = idf.d
            self.keys = keys
            self.previous_items = {}
            self.offset = offset if offset else []
            self(idf.aux_definition_params[idf.dict_var_name])

        def __call__(self, d: Optional[dict] = None):
            in_definition = d is not None
            d = self.d if d is None else d
            for offset_key in self.offset:
                d = d[offset_key]
            for k in self.keys:
                if in_definition and k in d:
                    self.previous_items[k] = d[k]
                del d[k]

        def revert(self) -> bool:
            d = self.d
            for offset_key in self.offset:
                if isinstance(d, list):
                    if not isinstance(offset_key, int) or offset_key >= len(d):
                        return False
                elif offset_key not in d:
                    return False
                d = d[offset_key]
            for k in self.keys:
                if k in self.previous_items:
                    d[k] = self.previous_items[k]
                else:
                    d[k] = None
            return True

    def delete_items(self, keys: list[str], offset: Optional[list[Union[str, int]]] = None):
        self.function_operations.append(self.DeleteItems(self, keys, offset=offset))

    # TODO: Fix conditional branch
    class ConditionalBranch(InvertibleOperation):

        def __init__(
                self,
                idf: 'InvertibleDictFunction',
                condition: any,
                condition_options: list[any],
                branches: list[Union['InvertibleDictFunction', 'InvertibleOperation']],
                else_branch: Optional[Union['InvertibleDictFunction', 'InvertibleOperation']] = None
        ):
            self.d = idf.d
            self.params = idf.params
            self.condition = condition
            self.condition_options = condition_options
            self.branches = branches
            if else_branch is not None:
                self.branches.append(else_branch)
                condition_options.append(None)

        def __call__(self):
            for condition_option, branch in zip(self.condition_options, self.branches):
                if self.params[self.condition] == condition_option:
                    branch()
                    return

        def revert(self) -> bool:
            d = self.d
            for condition_option, branch in zip(self.condition_options, self.branches):
                aux_d = deepcopy(d)
                if branch.revert(aux_d):
                    d.clear()
                    d.update(aux_d)
                    self.params[self.condition] = condition_option
                    return True
            return False

    def conditional_branch(
        self, condition: any, condition_options: list[any],
        branches: list[Union['InvertibleDictFunction', 'InvertibleOperation']],
        else_branch: Optional[Union['InvertibleDictFunction', 'InvertibleOperation']] = None
    ):
        self.function_operations.append(self.ConditionalBranch(
            self, condition, condition_options, branches, else_branch
        ))

    class SetFromParams(InvertibleOperation):
        def __init__(
            self, idf: 'InvertibleDictFunction',
            offset: list[Union[str, int]],
            key: str, value: any
        ):
            self.d = idf.d
            self.offset = offset
            self.params = idf.params
            self.key = key
            self.value = value
            self(idf.aux_definition_params[idf.dict_var_name])

        def __call__(self, d: Optional[dict] = None):
            d = self.d if d is None else d
            for offset_key in self.offset:
                d = d[offset_key]
            d[self.key] = self.value

        def revert(self) -> bool:
            d = self.d
            for offset_key in self.offset:
                if offset_key not in d:
                    return False
                d = d[offset_key]
            if self.key not in d:
                return False
            self.params[self.key] = d[self.key]
            del d[self.key]
            return True

    def set_from_params(self, offset: list[Union[str, int]], key: str, value: any):
        self.function_operations.append(self.SetFromParams(self, offset, key, value))

    class ParamsToParams(InvertibleOperation):
        def __init__(
            self, idf: 'InvertibleDictFunction',
            keys_dict: dict,
            replace: bool = False
        ):
            self.params = idf.params
            self.keys_dict = keys_dict
            self.previous_dict = deepcopy(idf.aux_definition_params)
            self.replace = replace
            self(idf.aux_definition_params)

        def __call__(self, params: Optional[dict] = None):
            if not params:
                params = self.params
            result = {new_key: params[old_key] for new_key, old_key in self.keys_dict.items()}
            if self.replace:
                params.clear()
            params.update(result)

        def revert(self) -> bool:
            result = deepcopy(self.previous_dict)
            for new_key, old_key in self.keys_dict.items():
                if new_key not in self.params:
                    return False
                result[old_key] = self.params[new_key]
            self.params.clear()
            self.params.update(result)
            return True

    def params_to_params(self, keys_dict: dict, replace: bool = False):
        self.function_operations.append(self.ParamsToParams(self, keys_dict, replace))

    class CustomOperation(InvertibleOperation):
        def __init__(
            self, idf: 'InvertibleDictFunction',
            func: callable, inverse_func: callable
        ):
            self.params = idf.params
            self.func = func
            self.inverse_func = inverse_func
            self(idf.aux_definition_params)

        def __call__(self, params: Optional[dict] = None):
            self.func(params if params else self.params)

        def revert(self) -> bool:
            return self.inverse_func(self.params)

    def custom_operation(self, func: callable, inverse_func: callable):
        self.function_operations.append(self.CustomOperation(self, func, inverse_func))

    def add_operations(self, operations: list[InvertibleOperation]):
        for operation in operations:
            self.function_operations.append(operation)

    def __call__(self, **initial_params) -> dict:
        for k, v in initial_params.items():
            if v is None:
                continue
            if k not in self.params:
                log_error(logger, f'Parameter {k} not found in', RuntimeError)
            elif k in [self.dict_var_name, self.diff_dict_var_name]:
                self.params[k].update(deepcopy(initial_params[k]))
            else:
                self.params[k] = initial_params[k]
        for operation in self.function_operations:
            operation()
        return self.params

    def revert(self, **ending_params) -> Optional[dict]:
        for k, v in ending_params.items():
            if k not in self.aux_definition_params:
                log_error(logger, f'Parameter {k} not found in the end execution params', RuntimeError)
            elif k in [self.dict_var_name, self.diff_dict_var_name]:
                self.params[k].update(deepcopy(ending_params[k]))
            else:
                self.params[k] = ending_params[k]
        for operation in reversed(self.function_operations):
            if not operation.revert():
                return None
        return self.params



# This file was generated automatically by scripts/user_classes_header_generator.py do not modify it!
# If the user access files are modified, this file has to be regenerated with the script.
from typing import Optional, Union
from pandas import DataFrame


class FilesLayerHeader:
    """
    This class is used to interact with the API at the file level.
    """

    def delete_file(
        self,
        uuid: Optional[str] = None,
        file_name: Optional[str] = None,
        with_shimoku_generated: bool = False,
    ):
        """
        Delete a file
        :param uuid: uuid of the file
        :param file_name: name of the file
        :param with_shimoku_generated: if True, allow to delete a file with the tag 'shimoku_generated'
        """
        pass

    def deleted_batched_dataframe(
        self,
        file_name: str,
        with_shimoku_generated: bool = False,
    ):
        """
        Delete a batched dataframe
        :param file_name: name of the file
        :param with_shimoku_generated: if True, allow to delete a file with the tag 'shimoku_generated'
        """
        pass

    def get_ai_model(
        self,
        uuid: Optional[str] = None,
        model_name: Optional[str] = None,
    ) -> any:
        """
        Get a model
        :param uuid: uuid of the file
        :param model_name: name of the model
        :return: model
        """
        pass

    def get_batched_dataframe(
        self,
        file_name: str,
    ) -> DataFrame:
        """
        Get a batched dataframe file group
        :param file_name: name of the file
        :return: dataframe
        """
        pass

    def get_dataframe(
        self,
        uuid: Optional[str] = None,
        file_name: Optional[str] = None,
    ) -> DataFrame:
        """
        Get a dataframe file
        :param uuid: uuid of the file
        :param file_name: name of the file
        :return: dataframe
        """
        pass

    def get_file_metadata(
        self,
        uuid: Optional[str] = None,
        file_name: Optional[str] = None,
    ) -> dict:
        """
        Get file metadata
        :param uuid: uuid of the file
        :param file_name: name of the file
        :return: the metadata
        """
        pass

    def get_object(
        self,
        uuid: Optional[str] = None,
        file_name: Optional[str] = None,
    ) -> bytes:
        """
        Get an object
        :param file_name: name of the file
        :param uuid: uuid of the file
        :return: object
        """
        pass

    def post_ai_model(
        self,
        model_name: str,
        model: callable,
        overwrite: bool = True,
        tags: Optional[list] = None,
        metadata: Optional[dict] = None,
    ):
        """
        Save a model
        :param model_name: name of the model
        :param model: model to be saved
        :param overwrite: if True, overwrite the file if it already exists
        :param tags: tags to be added to the file
        :param metadata: metadata to be added to the file
        """
        pass

    def post_batched_dataframe(
        self,
        file_name: str,
        df: DataFrame,
        batch_size: int = 10000,
        overwrite: bool = True,
        tags: Optional[list] = None,
        metadata: Optional[dict] = None,
    ):
        """
        Creates an multiple files saving a dataframe in batches, for big dataframes
        Uploads a dataframe in batches, for big dataframes
        :param file_name: name of the file
        :param df: dataframe to be saved
        :param batch_size: size of the batches
        :param overwrite: if True, overwrite the file if it already exists
        :param tags: tags to be added to the file
        :param metadata: metadata to be added to the file
        """
        pass

    def post_dataframe(
        self,
        file_name: str,
        df: DataFrame,
        overwrite: bool = True,
        tags: Optional[list] = None,
        metadata: Optional[dict] = None,
    ):
        """
        Save a dataframe
        :param df: dataframe to be saved
        :param file_name: name of the file
        :param overwrite: if True, overwrite the file if it already exists
        :param tags: tags to be added to the file
        :param metadata: metadata to be added to the file
        :return: object
        """
        pass

    def post_object(
        self,
        file_name: str,
        obj: bytes,
        overwrite: bool = True,
        tags: Optional[list] = None,
        metadata: Optional[dict] = None,
    ):
        """
        Save an object
        :param file_name: name of the file
        :param obj: object to be saved
        :param overwrite: if True, overwrite the file if it already exists
        :param tags: tags to be added to the file
        :param metadata: metadata to be added to the file
        """
        pass



"""
code modified from https://github.com/erdewit/nest_asyncio
This makes everything a little bit slower, there are other strategies, like threads, but this also has the advantage
of being able to run in one thread, which is useful for ipython notebooks and low-end devices.
Patch asyncio to allow nested event loops."""

import asyncio
import asyncio.events as events
import os
import sys
import threading
from contextlib import contextmanager, suppress
from heapq import heappop


def apply():
    """Patch asyncio to make its event loop reentrant."""
    _patch_asyncio()
    _patch_policy()
    _patch_task()

    try:
        loop = asyncio.get_event_loop()
        _patch_loop(loop)
    except (RuntimeError, ValueError):
        pass


def _patch_asyncio():
    """Patch asyncio module to use pure Python tasks and futures."""

    def run(main, *, debug=False):
        loop = asyncio.get_event_loop()
        loop.set_debug(debug)
        task = asyncio.ensure_future(main)
        try:
            return loop.run_until_complete(task)
        finally:
            if not task.done():
                task.cancel()
                with suppress(asyncio.CancelledError):
                    loop.run_until_complete(task)

    def _get_event_loop(stacklevel=3):
        loop = events._get_running_loop()
        if loop is None:
            loop = events.get_event_loop_policy().get_event_loop()
        return loop

    # Use module level _current_tasks, all_tasks and patch run method.
    if hasattr(asyncio, '_nest_patched'):
        return
    if sys.version_info >= (3, 6, 0):
        asyncio.Task = asyncio.tasks._CTask = asyncio.tasks.Task = \
            asyncio.tasks._PyTask
        asyncio.Future = asyncio.futures._CFuture = asyncio.futures.Future = \
            asyncio.futures._PyFuture
    if sys.version_info < (3, 7, 0):
        asyncio.tasks._current_tasks = asyncio.tasks.Task._current_tasks
        asyncio.all_tasks = asyncio.tasks.Task.all_tasks
    if sys.version_info >= (3, 9, 0):
        events._get_event_loop = events.get_event_loop = \
            asyncio.get_event_loop = _get_event_loop
    asyncio.run = run
    asyncio._nest_patched = True


def _patch_policy():
    """Patch the policy to always return a patched loop."""

    def get_event_loop(self):
        if self._local._loop is None:
            loop = self.new_event_loop()
            _patch_loop(loop)
            self.set_event_loop(loop)
        return self._local._loop

    policy = events.get_event_loop_policy()
    policy.__class__.get_event_loop = get_event_loop


def _patch_loop(loop):
    """Patch loop to make it reentrant."""

    def run_forever(self):
        with manage_run(self), manage_asyncgens(self):
            while True:
                self._run_once()
                if self._stopping:
                    break
        self._stopping = False

    def run_until_complete(self, future):
        with manage_run(self):
            f = asyncio.ensure_future(future, loop=self)
            if f is not future:
                f._log_destroy_pending = False
            while not f.done():
                self._run_once()
                if self._stopping:
                    break
            if not f.done():
                raise RuntimeError(
                    'Event loop stopped before Future completed.')
            return f.result()

    def _run_once(self):
        """
        Simplified re-implementation of asyncio's _run_once that
        runs handles as they become ready.
        """
        ready = self._ready
        scheduled = self._scheduled
        while scheduled and scheduled[0]._cancelled:
            heappop(scheduled)

        timeout = (
            0 if ready or self._stopping
            else min(max(
                scheduled[0]._when - self.time(), 0), 86400) if scheduled
            else None)
        event_list = self._selector.select(timeout)
        self._process_events(event_list)

        end_time = self.time() + self._clock_resolution
        while scheduled and scheduled[0]._when < end_time:
            handle = heappop(scheduled)
            ready.append(handle)

        for _ in range(len(ready)):
            if not ready:
                break
            handle = ready.popleft()
            if not handle._cancelled:
                handle._run()
        handle = None

    @contextmanager
    def manage_run(self):
        """Set up the loop for running."""
        self._check_closed()
        old_thread_id = self._thread_id
        old_running_loop = events._get_running_loop()
        try:
            self._thread_id = threading.get_ident()
            events._set_running_loop(self)
            self._num_runs_pending += 1
            if self._is_proactorloop:
                if self._self_reading_future is None:
                    self.call_soon(self._loop_self_reading)
            yield
        finally:
            self._thread_id = old_thread_id
            events._set_running_loop(old_running_loop)
            self._num_runs_pending -= 1
            if self._is_proactorloop:
                if (self._num_runs_pending == 0
                        and self._self_reading_future is not None):
                    ov = self._self_reading_future._ov
                    self._self_reading_future.cancel()
                    if ov is not None:
                        self._proactor._unregister(ov)
                    self._self_reading_future = None

    @contextmanager
    def manage_asyncgens(self):
        if not hasattr(sys, 'get_asyncgen_hooks'):
            # Python version is too old.
            return
        old_agen_hooks = sys.get_asyncgen_hooks()
        try:
            self._set_coroutine_origin_tracking(self._debug)
            if self._asyncgens is not None:
                sys.set_asyncgen_hooks(
                    firstiter=self._asyncgen_firstiter_hook,
                    finalizer=self._asyncgen_finalizer_hook)
            yield
        finally:
            self._set_coroutine_origin_tracking(False)
            if self._asyncgens is not None:
                sys.set_asyncgen_hooks(*old_agen_hooks)

    def _check_running(self):
        """Do not throw exception if loop is already running."""
        pass

    if hasattr(loop, '_nest_patched'):
        return
    if not isinstance(loop, asyncio.BaseEventLoop):
        raise ValueError('Can\'t patch loop of type %s' % type(loop))
    cls = loop.__class__
    cls.run_forever = run_forever
    cls.run_until_complete = run_until_complete
    cls._run_once = _run_once
    cls._check_running = _check_running
    cls._check_runnung = _check_running  # typo in Python 3.7 source
    cls._num_runs_pending = 0
    cls._is_proactorloop = (
        os.name == 'nt' and issubclass(cls, asyncio.ProactorEventLoop))
    if sys.version_info < (3, 7, 0):
        cls._set_coroutine_origin_tracking = cls._set_coroutine_wrapper
    cls._nest_patched = True


def _patch_task():
    """Patch the Task's step and enter/leave methods to make it reentrant."""

    def step(task, exc=None):
        curr_task = curr_tasks.get(task._loop)
        try:
            step_orig(task, exc)
        finally:
            if curr_task is None:
                curr_tasks.pop(task._loop, None)
            else:
                curr_tasks[task._loop] = curr_task

    Task = asyncio.Task
    if hasattr(Task, '_nest_patched'):
        return
    if sys.version_info >= (3, 7, 0):

        def enter_task(loop, task):
            curr_tasks[loop] = task

        def leave_task(loop, task):
            curr_tasks.pop(loop, None)

        asyncio.tasks._enter_task = enter_task
        asyncio.tasks._leave_task = leave_task
        curr_tasks = asyncio.tasks._current_tasks
        step_orig = Task._Task__step
        Task._Task__step = step
    else:
        curr_tasks = Task._current_tasks
        step_orig = Task._step
        Task._step = step
    Task._nest_patched = True





""""""


class APIError(Exception):
    def __init__(self, text, status_code=None):
        self.text = text
        self.status_code = status_code


class ResourceIdMissing(Exception):
    def __init__(self, text, status_code=None):
        self.text = text
        self.status_code = status_code


class CacheError(Exception):
    def __init__(self, text, status_code=None):
        self.text = text
        self.status_code = status_code


class TabsError(Exception):
    def __init__(self, text, status_code=None):
        self.text = text
        self.status_code = status_code


class TableError(Exception):
    def __init__(self, text, status_code=None):
        self.text = text
        self.status_code = status_code


class ModalError(Exception):
    def __init__(self, text, status_code=None):
        self.text = text
        self.status_code = status_code


class DataError(Exception):

    def __init__(self, text, status_code=None):
        self.text = text
        self.status_code = status_code


class FileError(Exception):

    def __init__(self, text, status_code=None):
        self.text = text
        self.status_code = status_code


class BentoboxError(Exception):
    def __init__(self, text, status_code=None):
        self.text = text
        self.status_code = status_code


class BoardError(Exception):
    def __init__(self, text, status_code=None):
        self.text = text
        self.status_code = status_code


class MenuPathError(Exception):
    def __init__(self, text, status_code=None):
        self.text = text
        self.status_code = status_code


class WorkspaceError(Exception):
    def __init__(self, text, status_code=None):
        self.text = text
        self.status_code = status_code


class ActivityError(Exception):
    def __init__(self, text, status_code=None):
        self.text = text
        self.status_code = status_code


class ActivityTemplateError(Exception):
    def __init__(self, text, status_code=None):
        self.text = text
        self.status_code = status_code


class AIFunctionError(Exception):
    def __init__(self, text, status_code=None):
        self.text = text
        self.status_code = status_code


class ShimokuFileError(Exception):

    def __init__(self, text, status_code=None):
        self.text = text
        self.status_code = status_code


class UniverseApiKeyError(Exception):

    def __init__(self, text, status_code=None):
        self.text = text
        self.status_code = status_code


class ActionError(Exception):
    def __init__(self, text, status_code=None):
        self.text = text
        self.status_code = status_code


class RoleError(Exception):
    def __init__(self, text, status_code=None):
        self.text = text
        self.status_code = status_code



import asyncio
from typing import Optional

from shimoku.api.resources.role import VALID_PERMISSIONS, VALID_RESOURCES, VALID_TARGETS

from shimoku.cli import CLIParser, CLIFuncParam
from shimoku.cli.cloud.cascade_get_resources import InitOptions, ResourceGetter, Universe
from shimoku.cli.utils import choose_from_menu, input_list, input_dict


def add_create_parser(parser: Optional[CLIParser] = None):
    """
    Function to add the create parser to a parser
    :param parser: Parser to add the create parser to
    :return: Create parser
    """
    params = {
        'name': 'create',
        'description': 'Commands to create resources'
    }
    if parser:
        create_parser = parser.add_command(**params)
    else:
        create_parser = CLIParser(**params)

    common_args = [
        CLIFuncParam(name='local-port', arg_type=int, arg_help='Local port to use', mandatory=False),
        CLIFuncParam(name='environment', arg_type=str, arg_help='Environment to use', mandatory=False),
        CLIFuncParam(name='access-token', arg_type=str, arg_help='Access token to use', mandatory=False),
        CLIFuncParam(name='universe-id', arg_type=str, arg_help='Universe ID to use', mandatory=False),
    ]

    module_functions = [
        universe_api_key,
        workspace,
        menu_path,
        board,
        board_menu_path_link,
        file,
        ai_input_file,
        role,
    ]

    for func in module_functions:
        create_parser.decor_add_command(common_args=common_args)(func)

    return create_parser


async def universe_api_key(
    description: str = CLIFuncParam(prompt=True),
    **kwargs
):
    """ Create a universe api key
    :param description: Description of the universe api key
    """
    universe: Universe = await ResourceGetter(InitOptions(**kwargs)).get_universe()
    await universe.create_universe_api_key(description=description)


async def workspace(
    name: str = CLIFuncParam(prompt=True),
    dont_create_default_roles: bool = CLIFuncParam(default=False, action='store_true', mandatory=False),
    theme: dict = CLIFuncParam(default=None, mandatory=False),
    **kwargs
):
    """ Create a workspace
    :param name: Workspace name to create
    :param dont_create_default_roles: Flag to not create default roles
    :param theme: Theme to use
    """
    businesses_layer = await ResourceGetter(InitOptions(**kwargs)).get_businesses_layer()
    await businesses_layer.create_workspace(
        name=name,
        create_default_roles=not dont_create_default_roles,
        theme=theme
    )


async def menu_path(
    workspace_id: Optional[str],
    name: str = CLIFuncParam(prompt=True),
    **kwargs
):
    """ Create a menu path in the workspace
    :param workspace_id: UUID of the workspace to use
    :param name: Menu path name to create
    :param order: Order of the menu path
    """
    resource_getter = ResourceGetter(InitOptions(workspace_id=workspace_id, **kwargs))
    apps_layer = await resource_getter.get_apps_layer()
    await apps_layer.get_menu_path(name=name)


async def board(
    workspace_id: Optional[str],
    order: Optional[int],
    name: str = CLIFuncParam(prompt=True),
    public: bool = CLIFuncParam(default=False, action='store_true', mandatory=False),
    not_enabled: bool = CLIFuncParam(default=False, action='store_true', mandatory=False),
    group_menu_paths: bool = CLIFuncParam(default=False, action='store_true', mandatory=False),
    **kwargs
):
    """ Create a board in the workspace
    :param workspace_id: UUID of the workspace to use
    :param name: Board name to create
    :param order: Order of the board
    :param public: Flag to set the board as public
    :param not_enabled: Flag to set the board as not enabled
    :param group_menu_paths: Flag to ask for menu paths to group
    """
    resource_getter = ResourceGetter(InitOptions(workspace_id=workspace_id, **kwargs))
    dashboards_layer = await resource_getter.get_dashboards_layer()
    dashboard = await dashboards_layer.create_board(
        name=name,
        order=order,
        is_public=public,
        is_disabled=not_enabled
    )
    if group_menu_paths:
        app_names = []
        business = await resource_getter.get_business()
        app_objs = await business.get_apps()
        options = [app['name'] for app in app_objs]+['']
        print("Select the menu paths to group in the board (empty to stop):")
        while True:
            chosen_app = choose_from_menu(options)
            if not chosen_app:
                break
            options.remove(chosen_app)
            app_names.append(chosen_app)

        await dashboards_layer.group_menu_paths(uuid=dashboard['id'], menu_path_names=app_names)


async def board_menu_path_link(
    workspace_id: Optional[str],
    board: str = CLIFuncParam(prompt=True),
    menu_path: str = CLIFuncParam(prompt=True),
    **kwargs
):
    """ Create a board menu path link
    :param workspace_id: UUID of the workspace to use
    :param board: Board id or name to link
    :param menu_path: Menu path to link
    """
    resource_getter = ResourceGetter(InitOptions(board=board, menu_path=menu_path, workspace_id=workspace_id, **kwargs))
    dashboards_layer = await resource_getter.get_dashboards_layer()
    dashboard, app = await asyncio.gather(
        resource_getter.get_dashboard(),
        resource_getter.get_app()
    )
    await dashboards_layer.add_menu_path_in_board(menu_path_id=app['id'], uuid=dashboard['id'])


def get_file_data(path_to_file: str) -> Optional[bytes]:
    """ Get the data of the file
    :param path_to_file: Path to the file
    """
    try:
        with open(path_to_file, 'rb') as file_:
            return file_.read()
    except FileNotFoundError:
        print(f"File {path_to_file} not found")


async def file(
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(prompt=True),
    name: str = CLIFuncParam(prompt=True),
    path_to_file: str = CLIFuncParam(prompt=True),
    ask_for_tags: bool = CLIFuncParam(default=False, action='store_true', mandatory=False),
    ask_for_metadata: bool = CLIFuncParam(default=False, action='store_true', mandatory=False),
    **kwargs
):
    """ Create a file in the menu path
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path id or name to use
    :param name: File name to create
    :param path_to_file: Path to the file to upload
    :param ask_for_tags: Flag to ask for tags
    :param ask_for_metadata: Flag to ask for metadata
    """
    tags, metadata = None, None
    file_content = get_file_data(path_to_file)
    if file_content is None:
        return
    resource_getter = ResourceGetter(InitOptions(menu_path=menu_path, workspace_id=workspace_id, **kwargs))
    files_layer = await resource_getter.get_files_layer()
    if ask_for_tags:
        tags = input_list()
    if ask_for_metadata:
        metadata = input_dict()
    await files_layer.post_object(file_name=name, obj=file_content, tags=tags, metadata=metadata)


async def ai_input_file(
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(prompt=True),
    name: str = CLIFuncParam(prompt=True),
    path_to_file: str = CLIFuncParam(prompt=True),
    **kwargs
):
    """ Create an ai input file in the menu path
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path id or name to use
    :param name: File name to create
    :param path_to_file: Path to the file to upload
    """
    file_content = get_file_data(path_to_file)
    if file_content is None:
        return
    resource_getter = ResourceGetter(InitOptions(menu_path=menu_path, workspace_id=workspace_id, **kwargs))
    ai_layer = await resource_getter.get_ai_layer()
    await ai_layer.create_input_files(input_files={name: file_content})


async def role(
    workspace_id: Optional[str],
    menu_path: str = CLIFuncParam(group='menu_path_or_board', mandatory=False),
    board: str = CLIFuncParam(group='menu_path_or_board', mandatory=False),
    name: str = CLIFuncParam(prompt=True),
    **kwargs
):
    """ Create a role in a specific level, leave menu path and board empty to create a workspace role
    and specify only one to create a menu path or board role respectively. Don't specify both.
    :param workspace_id: UUID of the workspace to use
    :param menu_path: Menu path id or name to use
    :param board: Board id or name to use
    :param name: Role name to create
    """
    role_resource = choose_from_menu(VALID_RESOURCES, 'Choose a resource: ')
    permission = choose_from_menu(VALID_PERMISSIONS, 'Choose a permission: ')
    target = choose_from_menu(VALID_TARGETS, 'Choose a target: ')
    resource_getter = ResourceGetter(InitOptions(
        workspace_id=workspace_id,
        menu_path=menu_path,
        board=board,
        **kwargs
    ))

    if menu_path:
        layer = await resource_getter.get_apps_layer()
        resource_id = (await resource_getter.get_app())['id']
    elif board:
        layer = await resource_getter.get_dashboards_layer()
        resource_id = (await resource_getter.get_dashboard())['id']
    else:
        layer = await resource_getter.get_businesses_layer()
        resource_id = (await resource_getter.get_business())['id']

    await layer.create_role(
        uuid=resource_id,
        role_name=name,
        resource=role_resource,
        permission=permission,
        target=target,
    )


if __name__ == "__main__":
    add_create_parser().parse_args()



from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from shimoku.code_gen.business_code_gen.apps_code_gen.code_gen_from_apps import AppCodeGen
    from shimoku.api.resources.report import Report


async def code_gen_from_iframe(
        self: 'AppCodeGen', report: 'Report'
) -> list[str]:
    """ Generate code for an iframe report.
    :param report: report to generate code from
    :return: list of code lines
    """
    code_lines = [
        'shimoku_client.plt.iframe(',
        f'    order={report["order"]},',
        f'    url="{report["dataFields"]["url"]}",',
    ]
    if report['dataFields']['height'] != 640:
        code_lines.append(f'    height={report["dataFields"]["height"]},')
    if report['sizeColumns'] != 12:
        code_lines.append(f'    cols_size={report["sizeColumns"]},')
    if report['sizePadding'] != '0,0,0,0':
        code_lines.append(f'    padding="{report["sizePadding"]}",')
    code_lines.append(')')
    return code_lines



# This file was generated automatically by scripts/user_classes_header_generator.py do not modify it!
# If the user access files are modified, this file has to be regenerated with the script.
from typing import Optional, Union
from pandas import DataFrame


class UniversesLayerHeader:
    """
    Class used to interact with the API at the universe level
    """

    def create_universe_api_key(
        self,
        uuid: str,
        description: str,
    ) -> dict:
        """
        Create a universe API key, this key has admin privileges.
        :param uuid: uuid of the universe
        :param description: description of the key
        """
        pass

    def delete_universe_api_key(
        self,
        uuid: str,
        api_key_uuid: str,
    ) -> bool:
        """
        Delete a universe API key.
        :param uuid: uuid of the universe
        :param api_key_uuid: uuid of the API key
        """
        pass

    def get_universe_activity_templates(
        self,
        uuid: str,
    ) -> list:
        """
        Get the universe activity templates as a list of dictionaries.
        :param uuid: uuid of the universe
        """
        pass

    def get_universe_api_keys(
        self,
        uuid: str,
    ) -> list:
        """
        Get the universe API keys.
        :param uuid: uuid of the universe
        """
        pass

    def get_universe_workspaces(
        self,
        uuid: str,
    ) -> list:
        """
        Get the universe workspaces.
        :param uuid: uuid of the universe
        """
        pass



import logging
from sys import stdout
from typing import Callable, Optional, Coroutine
from io import TextIOWrapper
from functools import wraps
from inspect import stack
import asyncio
from time import perf_counter
from abc import ABC

logger = logging.getLogger(__name__)


# Got this code from https://code.activestate.com/recipes/412603-stack-based-indentation-of-formatted-logging/
class IndentFormatter(logging.Formatter):
    """ Formatter that adds indentation to logging messages based on the stack. """
    def __init__(self, fmt=None, datefmt=None):
        logging.Formatter.__init__(self, fmt, datefmt)
        self.baseline = len(stack())

    def format(self, rec):
        """ Format the specified record as text. """
        _stack = stack()
        # we eliminate the unnecessary indents
        arrow = ('<- ' if 'Finished' in rec.msg else ('-> ' if 'Starting' in rec.msg else '| '))
        rec.indent = (' ' if rec.levelname == 'INFO' else '')+'｜ ' * (len(_stack) - self.baseline - 5) + arrow
        out = logging.Formatter.format(self, rec)
        del rec.indent
        return out


def log_error(logger, message, errorFunction):
    """ Log an error and raise an exception. """
    logger.error(message)
    raise errorFunction(message)


def logging_before_and_after(logging_level: Callable, name: Optional[str] = None) -> Callable:
    """ Decorator that logs before and after the execution of a function. """

    def decorator(func: Callable) -> Callable:

        def before_call(*args, **kwargs):
            """ Logs before the execution of the function. """
            enabled_for_debug = logging.root.isEnabledFor(logging.DEBUG)
            func_name = name
            if func_name is None:
                func_name = (func.__name__ if not enabled_for_debug else func.__qualname__)
            if 'logging_func_name' in kwargs:
                func_name = kwargs.pop('logging_func_name')

            underlined_text = '\033[4m' + func_name + '\033[0m'
            logging_level(
                f"Starting execution: {underlined_text}" +
                (f" with args: {args}, kwargs: {kwargs}" if enabled_for_debug else '')
            )
            initial_time = perf_counter()
            return initial_time, underlined_text

        def after_call(initial_time, underlined_text):
            """ Logs after the execution of the function. """
            time_spent = 1000*(perf_counter()-initial_time)
            logging_level(
                f"Finished execution: {underlined_text}, "
                f"elapsed time: {time_spent:.2f} ms")

        @wraps(func)
        async def awrapper(*args, **kwargs):
            """ Async version of the wrapper. """
            initial_time, underlined_text = before_call(*args, **kwargs)
            if 'logging_func_name' in kwargs:
                kwargs.pop('logging_func_name')
            result = await func(*args, **kwargs)
            after_call(initial_time, underlined_text)
            return result

        @wraps(func)
        def wrapper(*args, **kwargs):
            """ Normal version of the wrapper. """
            initial_time, underlined_text = before_call(*args, **kwargs)
            if 'logging_func_name' in kwargs:
                kwargs.pop('logging_func_name')
            result = func(*args, **kwargs)
            after_call(initial_time, underlined_text)
            return result

        return wrapper if not asyncio.iscoroutinefunction(func) else awrapper

    return decorator


def configure_logging(verbosity: Optional[str] = None, channel: Optional[TextIOWrapper] = stdout):
    """ Configures the logging module to use the specified verbosity and channel.
    :param verbosity: The verbosity level to use, can be 'DEBUG', 'INFO' or 'WARNING'.
    :param channel: The channel to use, can be a file or a stream.
    """

    verbosity = verbosity.upper()
    if verbosity not in ['DEBUG', 'INFO', 'WARNING']:
        raise ValueError(f"Invalid value for verbosity, only 'DEBUG' and 'INFO' are permitted, "
                         f"the value provided is: {verbosity}")

    if not hasattr(channel, 'write'):
        logger.warning("Specified channel doesn't have a 'write' method, using standard output")
        channel = stdout

    _format = f'%(asctime)s | %(levelname)s {"| %(indent)s%(name)s " if verbosity == "DEBUG" else ""}| %(message)s'
    # create a formatter and set its format and date format
    formatter = logging.Formatter(_format, datefmt='%Y-%m-%d %H:%M') if verbosity != 'DEBUG' \
        else IndentFormatter(_format, datefmt='%Y-%m-%d %H:%M')

    level = logging.DEBUG if verbosity == 'DEBUG' else (logging.INFO if verbosity == 'INFO' else logging.WARNING)
    # change the root logger
    logging.basicConfig(
        stream=channel,
        datefmt='%Y-%m-%d %H:%M',
        format=_format,
        level=level
    )

    # change all the logger handlers
    logging.getLogger().setLevel(level)

    for handler in logging.root.handlers:
        handler.setFormatter(formatter)
        handler.stream = channel


class ClassWithLogging(ABC):
    """
    This class is used to add logging to all the methods of a class.
    """
    _module_logger = logger
    _use_info_logging = False

    def __getattribute__(self, item):
        attr = object.__getattribute__(self, item)
        if callable(attr) and logger.isEnabledFor(logging.INFO):
            if not item.startswith('_') and self._use_info_logging:
                return logging_before_and_after(logging_level=self._module_logger.info, name=item)(attr)
            if logger.isEnabledFor(logging.DEBUG):
                return logging_before_and_after(logging_level=self._module_logger.debug, name=item)(attr)
        return attr



import subprocess
import socket
import psutil
import webbrowser
import time
import os

import logging
from shimoku.execution_logger import log_error
logger = logging.getLogger(__name__)


def check_server(server_host: str, server_port: int):
    """
    Check if the server is already running.
    :param server_host: The server host.
    :param server_port: The port where the server is running.
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    result = sock.connect_ex((server_host, server_port))
    sock.close()
    return result == 0


def kill_server(server_host: str, server_port: int):
    """
    Kill the server process.
    """
    cmd_line = ['local_server:app', '--host', server_host, '--port', str(server_port)]

    for process in psutil.process_iter(['pid', 'name', 'cmdline']):
        if process.info['cmdline'] is not None and (len(process.info['cmdline'])-2 == len(cmd_line) and
                                                    all([elm1 == elm2 for elm1, elm2 in
                                                         zip(cmd_line, process.info['cmdline'][2:])])):
            server_pid = None
            try:
                server_pid = process.info['pid']
                proc = psutil.Process(server_pid)
                proc.terminate()
                print(f"Server (PID: {server_pid}) terminated gracefully.")
            except psutil.NoSuchProcess:
                print(f"Process with PID {server_pid} not found.")


def create_server(environment: str, server_host: str, server_port: int, open_server_browser: bool):
    """
    Create a new server process and open the server URL in the default web browser.
    :param server_host: The server host.
    :param server_port: The port where the server will run.
    :param open_server_browser: Whether to open the server URL in the default web browser.
    """
    base_path = os.path.dirname(os.path.abspath(__file__))
    # Set the PYTHONPATH to include the directory of your FastAPI module
    env = os.environ.copy()
    env["PYTHONPATH"] = base_path + os.pathsep + env.get("PYTHONPATH", "")

    server_process = subprocess.Popen(
        ["uvicorn", "local_server:app", "--host", server_host, "--port", str(server_port)],
        env=env, stdout=subprocess.DEVNULL
    )

    print(f"Server started (PID: {server_process.pid})")

    max_time = 5
    # Wait for the server to fully initialize
    while not check_server(server_host, server_port):
        time.sleep(0.1)
        max_time -= 0.1
        if max_time <= 0:
            log_error(logger, "Server not initialized after 5 seconds.", RuntimeError)

    if open_server_browser:
        # Open the server URL in the default web browser
        webbrowser.open(f"http://{server_host}:{server_port}/graphql")

    environment_prefix = ''
    if environment != 'production':
        if environment.lower() in 'development':
            environment = 'develop'
        environment_prefix = f'{environment}.'

    webbrowser.open(f"http://{environment_prefix}shimoku.io/playground?port={server_port}")


def main():
    server_host = "127.0.0.1"
    server_port = 8000
    kill_server_flag = True
    server_running = check_server(server_host, server_port)
    if server_running:
        print("Server already running")
        if kill_server_flag:
            kill_server()
            server_running = False

    if not server_running:
        create_server(server_host, server_port, True)


if __name__ == "__main__":
    main()



from typing import List
from ..report import Report

import asyncio
import logging
logger = logging.getLogger(__name__)


class TabsGroup(Report):
    """ Tabs group report class """

    report_type = 'TABS'

    default_properties = dict(
        **Report.default_properties,
        tabs=dict(),
        sticky=False,
        variant='enclosedSolidRounded',
    )

    def __init__(self, *args, **kwargs):
        self.dirty = False
        super().__init__(*args, **kwargs)

    async def update(self, *args, **kwargs) -> bool:
        """ Update the tabs group on the server
        :return: True if the tabs group was updated, False otherwise
        """
        if not self.dirty:
            return False
        self.dirty = False
        await super().update()
        return True

    async def delete(self):
        """ Delete the tabs group from the server and all its children """
        delete_tasks = []
        for tab_dict in self['properties']['tabs'].values():
            for rd_id in tab_dict['reportIds']:
                delete_tasks.append(self._base_resource.parent.delete_component(rd_id))
        await asyncio.gather(*delete_tasks)
        await super().delete()

    def add_tab(self, tab: str):
        """ Add tab to the tabs group without saving it to the server
        :param tab: tab name
        """
        if tab in self['properties']['tabs']:
            return
        self['properties']['tabs'][tab] = {'order': len(self['properties']['tabs']), 'reportIds': []}
        self.dirty = True

    def add_report(self, tab: str, report: Report):
        """ Add report to the tabs group without saving it to the server
        :param tab: tab name
        :param report: report to add
        """
        self.add_tab(tab)

        if report['id'] in self['properties']['tabs'][tab]['reportIds']:
            logger.warning(f"Report {report['id']} already in tab {tab}")
            return

        self['properties']['tabs'][tab]['reportIds'].append(report['id'])
        self.dirty = True

    def remove_report(self, tab: str, report: Report):
        """ Remove report from the tabs group without saving it to the server
        :param tab: tab name
        :param report: report to remove
        """

        if tab not in self['properties']['tabs']:
            logger.warning(f"Tab {tab} not found")
            return

        if report['id'] not in self['properties']['tabs'][tab]['reportIds']:
            logger.warning(f"Report {report['id']} not in tab {tab}")
            return

        self['properties']['tabs'][tab]['reportIds'].remove(report['id'])
        self.dirty = True

    def change_tabs_order(self, tabs: List[str]):
        """ Change tabs order without saving it to the server
        :param tabs: list of tabs in the new order
        """
        all_tabs = list(self['properties']['tabs'].keys())

        for i, tab in enumerate(tabs):
            if tab not in self['properties']['tabs']:
                logger.warning(f"Tab {tab} not found")
                continue
            all_tabs.remove(tab)
            self['properties']['tabs'][tab]['order'] = i

        for i, tab in enumerate(all_tabs):
            self['properties']['tabs'][tab]['order'] = i + len(tabs)
        self.dirty = True

    def has_report(self, report: Report):
        """ Check if report is in the tabs group
        :param report: report to check
        """
        for tab in self['properties']['tabs'].values():
            if report['id'] in tab['reportIds']:
                return True
        return False

    def clear_content(self):
        """ Remove all reports from the tabs group without saving it to the server """
        self['properties']['tabs'] = dict()
        self.dirty = True



from shimoku.cli import CLIParser
import webbrowser
from typing import Optional
import psutil


def add_playground_parser(parser: Optional[CLIParser] = None):
    """
    Function to add the playground parser to a parser
    :param parser: Parser to add the playground parser to
    """

    params = {
        'name': 'playground',
        'description': 'Commands to interact with the playground',
    }

    if parser:
        playground_parser = parser.add_command(**params)
    else:
        playground_parser = CLIParser(**params)

    module_functions = [
        init,
        terminate
    ]

    for func in module_functions:
        playground_parser.decor_add_command()(func)

    return playground_parser


def init(local_port: Optional[int], env: Optional[str]):
    """ Start the playground server
    :param local_port: Port to use for the server
    :param env: Environment to use in the browser
    """
    from shimoku.playground.local_server import main
    if not local_port:
        local_port = 8000
    environment_prefix = f'{env}.' if env and env != 'production' else ''
    webbrowser.open(f"http://{environment_prefix}shimoku.io/playground?port={local_port}")
    main(server_port=local_port)


def terminate(local_port: Optional[int]):
    """ Terminate the playground server (for servers running on the background)
    :param local_port: Port to use for the server
    """
    if local_port is None:
        local_port = 8000
    cmd_lines = [['local_server:app', '--host', '127.0.0.1', '--port', str(local_port)], ['playground', 'init']]

    for process in psutil.process_iter(['pid', 'name', 'cmdline']):
        if process.info['cmdline'] is not None and any([all([elm in process.info['cmdline'] for elm in cmd_line])
                                                        for cmd_line in cmd_lines]):
            print(process.info['cmdline'])
            try:
                server_pid = process.info['pid']
                proc = psutil.Process(server_pid)
                proc.terminate()
                print(f"Server (PID: {server_pid}) terminated gracefully.")
            except psutil.NoSuchProcess:
                print(f"Process with PID {server_pid} not found.")



# This file was generated automatically by scripts/user_classes_header_generator.py do not modify it!
# If the user access files are modified, this file has to be regenerated with the script.
from typing import Optional, Union
from pandas import DataFrame


class MenuPathsLayerHeader:
    """
    This class is used to interact with the API at the Menu Path level.
    """

    def create_role(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        resource: Optional[str] = None,
        role_name: Optional[str] = None,
        permission: Optional[str] = None,
        target: Optional[str] = None,
    ) -> dict:
        """
    Create a new role at the resource level
    :param name: name of the resource
    :param uuid: uuid of the resource
    :param resource: resource of the role
    :param role_name: name of the role
    :param permission: permission of the role
    :param target: target of the role
    """
        pass

    def delete_all_menu_path_activities(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        with_linked_to_templates: Optional[bool] = False,
    ) -> bool:
        """
        Delete all activities of a menu path
        :param uuid: uuid of the menu path
        :param name: name of the menu path
        :param with_linked_to_templates: if True, delete all activities, even those linked to templates
        """
        pass

    def delete_all_menu_path_components(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        sub_path: Optional[str] = None,
    ) -> bool:
        """
        Delete all reports of a menu path
        :param uuid: uuid of the menu path
        :param name: name of the menu path
        :param sub_path: sub path of the menu path
        """
        pass

    def delete_all_menu_path_data_sets(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> bool:
        """
        Delete all datasets of a menu path
        :param uuid: uuid of the menu path
        :param name: name of the menu path
        """
        pass

    def delete_all_menu_path_files(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        with_shimoku_generated: Optional[bool] = False,
    ) -> bool:
        """
        Delete all files of an menu path
        :param uuid: uuid of the menu path
        :param name: name of the menu path
        :param with_shimoku_generated: whether the file is generated internally by the SDK
        """
        pass

    def delete_menu_path(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> bool:
        """
        Delete a menu path
        :param uuid: uuid of the menu path
        :param name:
        """
        pass

    def delete_role(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        role_id: Optional[str] = None,
        role_name: Optional[str] = None,
    ):
        """
    Delete the role at the resource level
    :param name: name of the resource
    :param uuid: uuid of the resource
    :param role_id: id of the role
    :param role_name: name of the role
    """
        pass

    def get_menu_path(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> dict:
        """
        Get a menu path
        :param uuid: uuid of the menu path
        :param name: name of the menu path
        :return: menu path metadata
        """
        pass

    def get_menu_path_activities(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> list:
        """
        Get the activities of a menu path
        :param uuid: uuid of the menu path
        :param name: name of the menu path
        """
        pass

    def get_menu_path_components(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> list:
        """
        Get the reports of a menu path
        :param uuid: uuid of the menu path
        :param name: name of the menu path
        """
        pass

    def get_menu_path_data_sets(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> list:
        """
        Get the datasets of a menu path
        :param uuid: uuid of the menu path
        :param name: name of the menu path
        """
        pass

    def get_menu_path_files(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        with_shimoku_generated: Optional[bool] = False,
    ) -> list:
        """
        Get the files of an menu path
        :param uuid: uuid of the menu path
        :param name: name of the menu path
        :param with_shimoku_generated: whether the file is generated internally by the SDK
        """
        pass

    def get_menu_path_sub_paths(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> list:
        """
        Get the path names of an menu path
        :param uuid: uuid of the menu path
        :param name: name of the menu path
        """
        pass

    def get_role(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        role_name: Optional[str] = None,
    ) -> list:
        """
    Get the role at the resource level
    :param name: name of the resource
    :param uuid: uuid of the resource
    :param role_name: name of the role
    """
        pass

    def get_roles(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> list:
        """
    Get the roles at the resource level
    :param name: name of the resource
    :param uuid: uuid of the resource
    """
        pass

    def update_menu_path(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        new_name: Optional[str] = None,
        hide_title: Optional[bool] = None,
        hide_path: Optional[bool] = None,
        show_breadcrumb: Optional[bool] = None,
        show_history_navigation: Optional[bool] = None,
        order: Optional[int] = None,
    ) -> bool:
        """
        Update a menu path
        :param uuid: uuid of the menu path
        :param name: the name of the menu path
        :param new_name: new name of the menu path
        :param hide_title: hide title of the menu path
        :param hide_path: hide path of the menu path
        :param show_breadcrumb: show breadcrumb of the menu path
        :param show_history_navigation: show history navigation of the menu path
        :param order: order of the menu path in the side menu
        :return: menu path metadata
        """
        pass



from typing import TYPE_CHECKING
from shimoku.code_gen.utils_code_gen import code_gen_from_list
if TYPE_CHECKING:
    from shimoku.code_gen.business_code_gen.apps_code_gen.code_gen_from_apps import AppCodeGen
    from shimoku.api.resources.report import Report


async def code_gen_from_form(
        self: 'AppCodeGen', report: 'Report', properties: dict
) -> list[str]:
    """ Generate code for a form report.
    :param report: report to generate code from
    :param properties: properties of the report
    :return: list of code lines
    """
    rds = (await report.get_report_data_sets())[0]
    rds_properties = rds['properties']
    form_groups = {group['title']: group['fields'] for group in rds_properties['fields']}
    input_form_params = []
    if rds_properties['fields'][0].get('nextFormGroup'):
        input_form_params.append(f'    dynamic_sequential_show=True,')
    if rds_properties.get('variant') == 'autoSend':
        input_form_params.append(f'    auto_send=True,')

    events_on_submit = []
    if 'events' not in properties:
        properties['events'] = {'onSubmit': []}
    for event in properties['events']['onSubmit']:
        if event['action'] == 'openModal':
            input_form_params.append(
                f'    modal="{(await self._app.get_report(event["params"]["modalId"]))["properties"]["hash"]}",'
            )
        elif event['action'] == 'openActivity':
            input_form_params.append(
                f'    activity_name="{(await self._app.get_activity(event["params"]["activityId"]))["name"]}",'
            )
        else:
            events_on_submit.append(event)
    if events_on_submit:
        code_gen_on_submit_events = code_gen_from_list(events_on_submit, 4)
        input_form_params.append(f'    on_submit_events={code_gen_on_submit_events[0][4:]}')
        input_form_params.extend(code_gen_on_submit_events[1:])

    report_params = [
        f'    order={report["order"]},',
    ]
    if report['sizeColumns'] != 12:
        report_params.append(f'    cols_size={report["sizeColumns"]},')
    if report['sizeRows'] != 1:
        report_params.append(f'    rows_size={report["sizeRows"]},')
    if report['sizePadding'] != '0,0,0,0':
        report_params.append(f'    padding="{report["sizePadding"]}",')
    return [
        'shimoku_client.plt.generate_input_form_groups(',
        f'    form_groups={form_groups},',
        *report_params,
        *input_form_params,
        ')'
    ]